# -*- coding: utf-8 -*-
"""
Classes and stuff that handle querying the interface for a passed in Orm class
"""
import copy
from collections import defaultdict
from collections.abc import AsyncIterable
import re

from datatypes import ListIterator

from .compat import *
from .utils import make_list, get_objects, make_dict


class Iterator(ListIterator, AsyncIterable):
    """The main iterator for all query methods that return iterators

    This is returned from the Query.get() method, this is also
    the Iterator class that is set in Orm.iterator_class

    fields --
        filter: callback, an iterator filter, all yielded rows will be passed
            through this callback and skipped if filter(row) returns False

    :example:
        # iterate through all the primary keys of some orm
        async for pk in SomeOrm.select_pk().query.get():
            print pk
    """
    @property
    def orm_class(self):
        return self.query.orm_class

    def __init__(self, cursor, query):
        """create an iterator for a query

        :param cursor: object, the cursor object retrieved from the interface
        :param query: Query, the query instance that produced this iterator
        """
        if query._filter:
            self.filter = query._filter

        self.query = query

        self._cursor = cursor
        self._cursor_exhausted = False
        self.field_names = self.query.fields_select.names()

    async def has_more(self):
        """Return true if there are more results for this query

        :returns: boolean, True if this query could've returned more results
        """
        ret = False
        if self.query.bounds.has_more():
            cursor = self._cursor
            # https://www.psycopg.org/docs/cursor.html#cursor.rowcount says
            # that future versions of the spec reserve the right to return None
            if cursor.rowcount == -1 or cursor.rowcount is None:
                try:
                    if await self[self.query.bounds.find_more_index()]:
                        ret = True

                except IndexError:
                    pass

            else:
                ret = self.query.bounds.limit_paginate == cursor.rowcount

        return ret

    async def close(self):
        """Close the cursor that's tied to this iterator"""
        await self._cursor.close()
        self._cursor_exhausted = True

    async def tolist(self):
        """Returns this iterator as a list

        This is just syntactic sugar around having to do:

            [r async for r in self]

        when a non asyncronous list is needed. It uses the same naming
        convention as array.tolist which also seems to be used by numpy and
        pandas also

        https://docs.python.org/3/library/array.html#array.array.tolist

        :returns: list, all the items in the iterator as a list
        """
        return [r async for r in self]

    def __iter__(self):
        """Make sure no one thinks we can iterate through this syncronously"""
        raise NotImplementedError()

    async def __aiter__(self):
        """
        https://docs.python.org/3/reference/datamodel.html#object.__aiter__
        """
        if self._cursor_exhausted:
            raise ValueError("Cursor has been exhausted, rerun the query")

        cursor_i = 0
        cursor_limit = -1
        if self.query.bounds.has_more():
            cursor_limit = self.query.bounds.limit

        try:
            async for row in self._cursor:
                if cursor_limit > 0 and cursor_i >= cursor_limit:
                    break

                o = self.hydrate(row)
                if self.filter(o):
                    yield o

                cursor_i += 1

        finally:
            await self.close()

    def __len__(self):
        """Make sure no one thinks we can get count syncronously"""
        raise NotImplementedError()

    async def count(self):
        """Get the number of rows this iterator represents

        Honestly, you probably shouldn't use this and instead do a count query
        instead since that gives you way more control

        :returns: int, the rows this iterator represents
        """
        cursor = self._cursor
        count = cursor.rowcount

        if count >= 0:
            bounds = self.query.bounds
            if bounds.has_more() and bounds.limit_paginate == count:
                # compensate for having pulled one extra row
                count -= 1

        else:
            # we couldn't get the rowcount from the cursor for some reason, so
            # we will need to query for it

            # SQLite cursor's will always have rowcount=-1 until the cursor is
            # exhausted or a fetch*() method has been called, ugh.
            # https://stackoverflow.com/a/839419/5006
            count = await self.query.copy().count()

        return count

    async def __getitem__(self, i):
        """Get a specific row or a slice of the rows represented by this
        iterator

        Honestly, you should probably make specific queries instead of using
        this since this will make new queries and you won't have control over
        things like the connection and stuff

        :param i: int|slice, the index(es) you want
        :returns: Orm|dict|Iterator
        """
        q = self.query.copy()
        b = q.bounds
        limit = b.limit if b.has_limit() else await self.count()
        b = q.bounds_class(limit=limit, offset=q.bounds.offset)

        if isinstance(i, slice):
            if i.step:
                raise ValueError("slice stepping is not supported")

            if i.start:
                start = b.find_offset(i.start)

            else:
                start = b.offset

            if i.stop:
                stop = b.find_offset(i.stop)

            else:
                stop = b.limit

            q.limit(stop - start).offset(start)
            return type(self)(await q.cursor(), q)

        else:
            offset = b.find_offset(i)

            o = await q.offset(offset).one()
            if o is None:
                raise IndexError("Iterator index {} out of range".format(i))

            return o

    def __repr__(self):
        format_str = "[ ... {} ... ]"
        format_args = [self.__class__.__name__]

        orm_class = self.orm_class
        if orm_class:
            format_str = "[ ... {} on {} ... ]"
            format_args.append(orm_class.__name__)

        return format_str.format(*format_args)

    def filter(self, o):
        """run o through the filter, if True then orm o should be included

        NOTE -- The filter callback needs to account for non Orm instance
            values of o

        :param o: Orm|tuple[Any]|Any, usually an Orm instance but can also be
            a tuple or single value
        :returns: boolean, True if o should be filtered
        """
        return True

    def hydrate(self, d):
        """Prepare the raw dict d returned from the interface cursor to be
        returned by higher level objects, this will usually mean hydrating an
        Orm instance or stuff like that

        NOTE -- this will call Orm.hydrate and should be the only place that
            will call that method

        :param d: dict, the raw dict cursor result returned from the interface
        :returns: Orm|tuple[Any]|Any, usually an Orm instance populated with d
            but can also be a tuple if the query selected more than one field.
            If the query selected one field then just that value will be
            returned
        """
        r = None
        orm_class = self.orm_class
        if self.field_names:
            field_vals = []
            for field_name in self.field_names:
                fv = d[field_name]
                if orm_class:
                    fv = orm_class.schema.fields[field_name].iget(None, fv)
                field_vals.append(fv)
            r = field_vals if len(self.field_names) > 1 else field_vals[0]

        else:
            if orm_class:
                r = orm_class.hydrate(d)

            else:
                r = d

        return r


class QueryBounds(object):
    """Stores the bounds (eg, DESC, ASC) information for a Query. It has a lot
    of hooks to make setting limit and offset easier
    """
    @property
    def limit(self):
        l = self._limit
        return l if l else 0

    @limit.setter
    def limit(self, v):
        v = int(v)
        if v < 0:
            raise ValueError("Limit cannot be negative")
        self._limit = v

    @limit.deleter
    def limit(self):
        self._limit = None

    @property
    def limit_paginate(self):
        limit = 0 if self._limit is None else self._limit
        return limit + 1 if limit > 0 else 0

    @property
    def offset(self):
        offset = self._offset
        if offset is None:
            page = self.page
            #limit = self.limit
            limit = self._limit
            if not limit: limit = 1
            offset = (page - 1) * limit

        else:
            offset = offset if offset >= 0 else 0

        return offset

    @offset.setter
    def offset(self, v):
        v = int(v)
        if v < 0:
            raise ValueError("Offset cannot be negative")
        del self.page
        self._offset = v

    @offset.deleter
    def offset(self):
        self._offset = None

    @property
    def page(self):
        page = 0 if self._page is None else self._page
        return page if page >= 1 else 1

    @page.setter
    def page(self, v):
        v = int(v)
        if v < 0:
            raise ValueError("Page cannot be negative")
        del self.offset
        self._page = int(v)

    @page.deleter
    def page(self):
        self._page = None

    def __init__(self, limit=None, page=None, offset=None):
        self.paginate = False
        self._limit = limit
        self._offset = offset
        self._page = page

    def set(self, limit=None, page=None, offset=None):
        if limit is not None:
            self.limit = limit

        if page is not None:
            if offset:
                raise ValueError("Cannot pass in both offset and page")
            self.page = page

        if offset is not None:
            self.offset = offset

    def get(self):
        limit = self.limit_paginate if self.paginate else self.limit
        return (limit, self.offset)

    def __bool__(self):
        return self.limit > 0 or self.offset > 0

    def has(self):
        return bool(self)

    def has_limit(self):
        return self.limit > 0

    def has_more(self):
        """Returns True if the current bounds are set up to query one extra row
        in order to make pagination easier (to know if there should be a next
        link)

        :returns: boolean
        """
        return self.has_limit() and self.paginate

    def has_pages(self):
        return self.has_more()

    def is_paginated(self):
        return self.has_more()

    def __str__(self):
        return "limit: {}, offset: {}".format(self.limit, self.offset)

    def find_offset(self, i):
        """Given an index i, use the current offset and limit to find the
        correct offset i would be

        :param i: int, the index used to find the new offset
        :returns: int, the new offset
        """
        if i >= 0:
            offset = self.offset + i
            if self.has_limit():
                maximum_offset = self.limit + self.offset
                if offset > maximum_offset:
                    raise IndexError(
                        "Iterator index {} out of range".format(i)
                    )

        else:
            limit = self.limit
            offset = limit + i

            if offset < 0:
                raise IndexError("Iterator index {} out of range".format(i))

        return offset

    def find_more_index(self):
        #return self.offset + self.limit_paginate
        return self.offset + self.limit


class QueryField(object):
    """Holds information for a field in the query"""
    @property
    def schema(self):
        return self.query.schema if self.query else None

    def __init__(self, query, field_name, field_val=None, **kwargs):
        self.query = query
        self.operator = kwargs.pop("operator", None)
        self.is_list = kwargs.pop("is_list", False)
        self.direction = kwargs.pop("direction", None) # 1 = ASC, -1 = DESC
        self.increment = kwargs.pop("increment", False) # Query.incr_field
        self.raw = kwargs.pop("raw", False)
        self.clause = kwargs.pop("clause", "")
        self.function_name = kwargs.pop("function_name", "")
        self.or_clause = False
        self.kwargs = kwargs

        if self.raw:
            self.name = field_name
            self.value = field_val

        else:
            self.set_name(field_name)
            self.set_value(field_val)

    def set_name(self, field_name):
        field_name, function_name = self.parse(field_name, self.schema)

        self.name = field_name

        if function_name:
            self.function_name = function_name

        if self.function_name and self.in_select_clause():
            self.alias = field_name

    def set_value(self, field_val):
        # we set this here so things like .is_subquery work as expected in a
        # Field.iquery method body
        self.value = field_val

        if self.is_list and not self.is_subquery():
            if field_val:
                field_val = make_list(field_val)
                for i in range(len(field_val)):
                    field_val[i] = self.iquery(field_val[i])

        else:
            field_val = self.iquery(field_val)

        self.value = field_val

    def iquery(self, field_val):
        schema = self.schema
        if schema:
            schema_field = getattr(schema, self.name)
            if ref_class := schema_field.ref_class:
                if isinstance(field_val, ref_class):
                    field_val = field_val.pk

            field_val = schema_field.iquery(self, field_val)

        return field_val

    def parse(self, field_name, schema):
        function_name = ""
        if schema:
            function_name = ""
            m = re.match(r"^([^\(]+)\(([^\)]+)\)$", field_name)
            if m:
                function_name = m.group(1)
                field_name = m.group(2)

            try:
                field_name = schema.field_name(field_name)

            except AttributeError:
                field_name = schema.field_model_name(field_name)

        return field_name, function_name

    def in_clause(self, clause):
        """Check if clause corresponds to self.clause

        :param clause: str, the clause name (eg, where, select, sort, set)
        :returns: bool, True if the clauses match
        """
        return self.clause == clause

    def in_select_clause(self):
        return self.in_clause("select")

    def in_set_clause(self):
        """Return True if this field instance belongs to the fields_set
        clause/portion of the query"""
        return self.in_clause("set")

    def in_where_clause(self):
        """Return True if this field instance belongs to the fields_where
        clause/portion of the query"""
        return self.in_clause("where")

    def is_subquery(self):
        """Return True if this field's value is a subquery"""
        return isinstance(self.value, Query)


class QueryFields(list):
    """Holds all the QueryField instances for a given clause"""
    @property
    def fields(self):
        """Returns a dict of field_name: field_value"""
        ret = {}
        for f in self:
            ret[f.name] = f.value

        return ret

    def __init__(self):
        self.clear()

    def names(self):
        """Return all the field names in the order they were first seen"""
        ret = []
        seen = set()
        for f in self:
            if f.name not in seen:
                ret.append(f.name)
                seen.add(f.name)
        return ret

    def append(self, field):
        index = len(self)
        super().append(field)
        self.field_names[field.name].append(index)

    def get_field(self, field_name):
        """Return the last field information for field_name

        This is handy for getting set information. If you need all the set
        values for field_name then you would just use
        self.field_names[field_name]

        :param field_name: str, the field name you're looking for
        :returns: QueryField|None
        """
        try:
            return self[self.field_names[field_name][-1]]

        except KeyError:
            return None

    def __contains__(self, field_name):
        return field_name in self.field_names

    def __setitem__(self, *args, **kwargs):
        raise NotImplementedError()
    __delitem__ = __setitem__

    def clear(self):
        self.field_names = defaultdict(list)
        self.options = {}
        super().clear()

    def todict(self):
        """Returns a dict of field_name: QueryField instance"""
        ret = {}
        for f in self:
            ret[f.name] = f

        return ret


class Query(AsyncIterable):
    """Handle standard query creation and allow interface querying

    This is the glue between Orm and Interface. For the most part you will
    never interact with the Interface directly but instead use Orm and Query,
    and you won't usually ever instantiate this class directly but instead use
    the Orm.query magic class property

    :example:
        q = Query(orm_class)
        q.select_bar().select_che().eq_foo(1).desc_bar().limit(10).page(2)
        q.get()
        # SELECT
        #   bar,
        #   che
        # FROM
        #   <orm_class.table_name>
        # WHERE
        #   foo = 1
        # ORDER BY bar DESC
        # LIMIT 10 OFFSET 20
    """
    field_class = QueryField
    fields_set_class = QueryFields
    fields_select_class = QueryFields
    fields_where_class = QueryFields
    fields_sort_class = QueryFields
    bounds_class = QueryBounds
    iterator_class = Iterator

    @property
    def interface(self):
        return self.orm_class.interface if self.orm_class else None

    @property
    def schema(self):
        return self.orm_class.schema if self.orm_class else None

    @property
    def schemas(self):
        """Find and return all the schemas that are needed for this query to
        complete successfully

        Another way to put this is all the schemas this query touches

        :returns: list, a list of Schema instances
        """
        schemas = []
        s = self.schema
        if s:
            schemas.append(s)

        for f in self.fields_where:
            if isinstance(f.value, Query):
                s = f.value.schema
                if s:
                    schemas.append(s)

        return schemas

    @property
    def OR(self):
        """Wraps left and right field statements with an OR clause, you can
        chain as many OR calls as you want to create an any length OR clause

        I don't love that I had to use OR instead of "or" but "or" is a
        reserved keyword and I thought OR was better than like _or, or_ or _or_

        :Example:
            self.eq_foo(1).OR.eq_foo(5).OR.eq_foo(10)
            # (foo=1 OR foo=5 OR foo=10)
        """
        self.fields_where[-1].or_clause = True
        return self

    @property
    def or_(self): return self.OR
    @property
    def _or(self): return self.OR
    @property
    def _or_(self): return self.OR

    @property
    def AND(self):
        """This is just here for completeness with .OR since, by default, any
        statements will be joined by AND"""
        self.fields_where[-1].or_clause = False
        return self

    @property
    def and_(self): return self.AND
    @property
    def _and(self): return self.AND
    @property
    def _and_(self): return self.AND

    def __init__(self, orm_class=None, **kwargs):
        # needed to use the db querying methods like get(), if you just want to
        # build a query then you don't need to bother passing this in
        self.orm_class = orm_class
        self.reset()

    def reset(self):
        self._filter = None
        self.fields_set = self.fields_set_class()
        self.fields_select = self.fields_select_class()
        self.fields_where = self.fields_where_class()
        self.fields_sort = self.fields_sort_class()
        self.bounds = self.bounds_class()
        self.compounds = []

    def ref(self, orm_classpath):
        """
        takes a classpath to allow query-ing from another Orm class

        the reason why it takes string paths is to avoid infinite recursion
        import problems because an orm class from module A might have a ref
        from module B and sometimes it is handy to have module B be able to get
        the objects from module A that correspond to the object in module B,
        but you can't import module A into module B because module B already
        imports module A.

        :param orm_classpath: string|type, a full python class path (eg,
            foo.bar.Che) or an actual model.Orm python class
        return -- Query
        """
        # split orm from module path
        if isinstance(orm_classpath, basestring):
            orm_module, orm_class = get_objects(orm_classpath)

        else:
            orm_class = orm_classpath

        return orm_class.query

    def ref_class(self, orm_classpath):
        """alias to .ref, here to match Orm.ref_class syntax"""
        return self.ref(orm_classpath)

    def find_methods(self, method_name):
        """Given a method name like <OPERATOR>_<FIELD_NAME> or 
        <FIELD_NAME>_<OPERATOR>, split those into <OPERATOR> and <FIELD_NAME>
        if there is an existing <OPERATOR>_field method that exists

        :example:
            self.gt_foo() # (<gt_field>, None, "foo")

        :returns: tuple[str, str], (<FIELD_METHOD>, <FIELD_NAME>)
        """
        # infinite recursion check, if a *_field method gets in here then it
        # doesn't exist
        if method_name.endswith("_field"):
            raise AttributeError(method_name)

        try:
            # check for <NAME>_<FIELD_NAME>
            name, field_name = method_name.split("_", 1)

        except ValueError:
            raise AttributeError("invalid potential method: {}".format(
                method_name
            ))

        else:
            if not name:
                raise AttributeError(
                    'Could not resolve methods from {}"'.format(
                        method_name
                    )
                )

            else:
                field_method_name = "{}_field".format(name)
                field_method = getattr(self, field_method_name, None)
                if not field_method:
                    # let's try reversing the split, so <FIELD_NAME>_<NAME>
                    field_name, name = method_name.rsplit("_", 1)
                    field_method_name = "{}_field".format(name)
                    field_method = getattr(self, field_method_name, None)

                if not field_method:
                    raise AttributeError(
                        'No field method derived from {}'.format(
                            method_name
                            )
                        )

        return field_method, field_name

    def __getattr__(self, method_name):
        """Allows fluid interface of .<OPERATOR>_<FIELD_NAME>(<FIELD_VALUE>)

        :returns: callback, this will return a callback that wraps a 
            *_field method and passes in <FIELD_NAME> to the *_field method
        """
        field_method, field_name = self.find_methods(method_name)

        def callback(*args, **kwargs):
            if field_method:
                ret = field_method(field_name, *args, **kwargs)

            return ret

        return callback

    def create_field(self, field_name, field_val=None, **kwargs):
        """Creates a QueryField instance"""
        f = self.field_class(self, field_name, field_val, **kwargs)
        return f

    def create_iterator(self, cursor):
        """Creates a query.Iterator instance wrapping cursor

        :param cursor: object, this is retrieved from the Interface
        :returns: Iterator
        """
        if self.orm_class:
            iterator = self.orm_class.iterator_class(cursor, query=self)

        else:
            iterator = self.iterator_class(cursor, query=self)

        return iterator

    def append_compound(self, operator, queries, **kwargs):
        """Internal method used by .intersect(), .union(), and .difference()"""
        for i, query in enumerate(queries):
            if self.fields_select:
                if not query.fields_select:
                    if self.schema == query.schema:
                        query.select(*self.fields_select.names())

            else:
                if i == 0 and query.fields_select:
                    self.select(*query.fields_select.names())

        self.compounds.append((operator, queries))
        return self

    def append_operation(self, operator, field_name, field_val=None, **kwargs):
        """Internal method that appends an operation to the where clause"""
        kwargs["operator"] = operator
        kwargs["clause"] = "where"
        f = self.create_field(field_name, field_val, **kwargs)
        self.fields_where.append(f)
        return self

    def append_sort(self, direction, field_name, field_val=None, **kwargs):
        """Internal method that appends a sort to the sort clause

        sort this query by field_name in direction

        used to be named sort_field

        :param field_name: string, the field to sort on
        :param direction: integer, negative for DESC, positive for ASC
        :param field_val: list, the order the rows should be returned in
        """
        if direction > 0:
            direction = 1

        elif direction < 0:
            direction = -1

        else:
            raise ValueError("direction {} is undefined".format(direction))

        kwargs["direction"] = direction
        kwargs["is_list"] = True
        kwargs["clause"] = "sort"
        f = self.create_field(field_name, field_val, **kwargs)
        self.fields_sort.append(f)
        return self

    def distinct(self, *field_names, **kwargs):
        """Mark the passed in query select fields as distinct"""
        self.fields_select.options["distinct"] = True
        return self.select(*field_names, **kwargs)

    def select_field(self, field_name, **kwargs):
        """set a field to be selected, this is automatically called when you do
        select_FIELDNAME(...)"""
        if field_name == "*":
            self.fields_select.options["all"] = True

        else:
            kwargs["clause"] = "select"
            field = self.create_field(field_name, **kwargs)
            self.fields_select.append(field)

        return self

    def select(self, *field_names, **kwargs):
        """set multiple fields to be selected, this is the many version of 
        .select_field"""
        for field_name in make_list(field_names):
            self.select_field(field_name, **kwargs)
        return self

    def set_field(self, field_name, field_val, **kwargs):
        """Set a field into .fields_set attribute

        In insert/update queries, these are the fields that will be
        inserted/updated into the db
        """
        kwargs["clause"] = "set"
        field = self.create_field(field_name, field_val, **kwargs)
        self.fields_set.append(field)
        return self

    def set(self, fields=None, **fields_kwargs):
        """completely replaces the current .fields with fields and fields_kwargs
        combined, this is the many version of .set_field
        """
        fields = make_dict(fields, fields_kwargs)
        for field_name, field_val in fields.items():
            self.set_field(field_name, field_val)
        return self

    def incr_field(self, field_name, increment=1, field_val=None, **kwargs):
        """Set a field to be incremented into .fields_set attribute

        In update queries, these are the fields that will be
        updated into the db using field_name = field_name + field_val
        for an atomic increment
        """
        return self.set_field(
            field_name,
            field_val,
            increment=increment,
            **kwargs
        )

    def intersect(self, *queries, **kwargs):
        """Intersect a set of queries. Returns rows that are common in all the 
        queries.

        Compound Query instances will only return values, not Orm instances

        :Example:
            Query().intersect(
                Query().select_foo(),
                Query().select_foo(),
            ).all()

        https://www.sqlite.org/syntax/compound-operator.html

        :param *queries: two or more queries, they should all select the same
            field types, the first query sets the name of the return columns
        """
        return self.append_compound("intersect", queries, **kwargs)

    def union(self, *queries, **kwargs):
        """Return all the rows from all the queries

        See .intersect() since this works very similar
        """
        return self.append_compound("union", queries, **kwargs)

    def difference(self, *queries, **kwargs):
        """Return everything in the first query that isn't in any of the other
        queries.

        Return the rows with rows in the first query that are not in the other
        queries

        This is named .difference() instead of .except() because except is a
        python keyword.

        See .intersect() since this works very similar
        """
        return self.append_compound("except", queries, **kwargs)

    def eq_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> = <FIELD_VALUE>"""
        return self.append_operation(
            "eq",
            field_name,
            field_val,
            **field_kwargs
        )

    def ne_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> != <FIELD_VALUE>"""
        return self.append_operation(
            "ne",
            field_name,
            field_val,
            **field_kwargs
        )

    def between_field(self, field_name, low, high):
        """<FIELD_NAME> >= low AND <FIELD_NAME> <= high"""
        self.gte_field(field_name, low)
        self.lte_field(field_name, high)
        return self

    def lte_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> <= <FIELD_VALUE>"""
        return self.append_operation(
            "lte",
            field_name,
            field_val,
            **field_kwargs
        )

    def lt_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> < <FIELD_VALUE>"""
        return self.append_operation(
            "lt",
            field_name,
            field_val,
            **field_kwargs
        )

    def gte_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> >= <FIELD_VALUE>"""
        return self.append_operation(
            "gte",
            field_name,
            field_val,
            **field_kwargs
        )

    def gt_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> > <FIELD_VALUE>"""
        return self.append_operation(
            "gt",
            field_name,
            field_val,
            **field_kwargs
        )

    def in_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> IN (<FIELD_VALUE>)

        :param field_val: list, a list of field_val values
        """
        field_kwargs["is_list"] = True
        return self.append_operation(
            "in",
            field_name,
            field_val,
            **field_kwargs
        )

    def nin_field(self, field_name, field_val=None, **field_kwargs):
        """<FIELD_NAME> NOT IN (<FIELD_VALUE>)

        :param field_val: list, a list of field_val values
        """
        field_kwargs["is_list"] = True
        return self.append_operation(
            "nin",
            field_name,
            field_val,
            **field_kwargs
        )

    def startswith_field(self, field_name, field_val, **field_kwargs):
        """<FIELD_NAME> LIKE <FIELD_VALUE>%"""
        return self.like_field(
            field_name,
            "{}%".format(field_val),
            **field_kwargs
        )

    def endswith_field(self, field_name, field_val, **field_kwargs):
        """<FIELD_NAME> LIKE %<FIELD_VALUE>"""
        return self.like_field(
            field_name,
            "%{}".format(field_val),
            **field_kwargs
        )

    def contains_field(self, field_name, field_val, **field_kwargs):
        """<FIELD_NAME> LIKE %<FIELD_VALUE>%"""
        return self.like_field(
            field_name,
            "%{}%".format(field_val),
            **field_kwargs
        )

    def like_field(self, field_name, field_val, **field_kwargs):
        """Perform a field_name LIKE field_val query

        :param field_name: string, the field we are filtering on
        :param field_val: string, the like query: %val, %val%, val%
        :returns: self, for fluid interface
        """
        if not field_val:
            raise ValueError("Cannot LIKE nothing")

        return self.append_operation(
            "like",
            field_name,
            field_val,
            **field_kwargs
        )

    def nlike_field(self, field_name, field_val, **field_kwargs):
        """Perform a field_name NOT LIKE field_val query

        :param field_name: string, the field we are filtering on
        :param field_val: string, the like query: %val, %val%, val%
        :returns: self, for fluid interface
        """
        if not field_val:
            raise ValueError("Cannot NOT LIKE nothing")

        return self.append_operation(
            "nlike",
            field_name,
            field_val,
            **field_kwargs
        )

    def raw_field(self, field_name, *field_vals, **field_kwargs):
        field_kwargs["raw"] = True
        return self.append_operation(
            "raw",
            field_name,
            field_vals,
            **field_kwargs
        )

    def asc(self, *field_names):
        """the many version of .asc_field"""
        for field_name in field_names:
            self.asc_field(field_name)
        return self

    def asc_field(self, field_name, field_val=None):
        """<FIELD_NAME> ASC"""
        return self.append_sort(1, field_name, field_val)

    def desc(self, *field_names):
        """the many version of .desc_field"""
        for field_name in field_names:
            self.desc_field(field_name)
        return self

    def desc_field(self, field_name, field_val=None):
        """<FIELD_NAME> DESC"""
        return self.append_sort(-1, field_name, field_val)

    def filter(self, predicate):
        """Set the predicate (callback) for an iterator returned from this
        instance

        :param predicate: callable, this will be passed to the iterator in the
            .create_iterator method
        :returns: self, for fluid interface
        """
        self._filter = predicate
        return self

    def limit(self, limit):
        """LIMIT <limit>"""
        self.bounds.limit = limit
        return self

    def offset(self, offset):
        """OFFSET <offset>"""
        self.bounds.offset = offset
        return self

    def page(self, page):
        """OFFSET <PAGE_CONVERTED_TO_OFFSET>"""
        self.bounds.page = page
        return self

    def copy(self):
        """nice handy wrapper around the deepcopy"""
        return copy.deepcopy(self)

    def __deepcopy__(self, memodict={}):
        instance = type(self)(self.orm_class)
        ignore_keys = set(["_interface", "interface"])
        for key, val in self.__dict__.items():
            if key not in ignore_keys:
                setattr(instance, key, copy.deepcopy(val, memodict))
        return instance

    def __str__(self):
        """Attempt to render the query, this should be used mainly for debugging
        and not in logs or things like that since this can be intense and
        multiline and also wrong
        """
        try:
            return self.render()

        except AttributeError:
            return super().__str__()

    def render(self, **kwargs):
        """Render the query

        :returns: string, the rendered query, this is not assured to be a valid
            query but is handy for quickly debugging what the query roughly
            looks like
        """
        return self.interface.render(self.schema, self, **kwargs)

    async def __aiter__(self):
        """Allows fluid interface for iterating through the query

        :Example:
            [r async for r in q]
        """
        async for row in await self.get():
            yield row

    async def tolist(self, **kwargs):
        """Wraps the Iterator.tolist method and does the same thing"""
        iterator = await self.get(**kwargs)
        return await iterator.tolist()

    async def raw(self, query_str, *query_args, **kwargs):
        """Send a raw query to the interface without any processing

        NOTE -- This will allow you to make any raw query and will usually
        return raw results

        :param query_str: str, the raw query for whatever the backend interface
            is
        :param query_args: list, if you have named parameters, you can pass in
            the values
        :param **kwargs: passed directly to the interface
        :returns: Any, depends on the interface and the type of query
        """
        return await self.interface.raw(query_str, *query_args, **kwargs)

    async def cursor(self, **kwargs):
        """Return a raw Interface cursor

        :param **kwargs: passed directly to the interface
        :returns: object, a cursor instance that can retrieve results from the
            interface's backend
        """
        return await self.interface.get(
            self.schema,
            self,
            cursor_result=True,
            **kwargs
        )

    async def get(self, **kwargs):
        """
        get results from the db

        :returns: Iterator
        """
        self.bounds.paginate = kwargs.pop("paginate", True)
        cursor = await self.cursor(**kwargs)
        return self.create_iterator(cursor)

    async def one(self, **kwargs):
        """get one row from the db"""
        ret = None

        kwargs["paginate"] = False
        async for ret in await self.limit(1).get(**kwargs):
            break

        return ret

    async def count(self, **kwargs):
        """return the row count of the criteria

        :param **kwargs: passed through to the interface
        :returns: int, the row count
        """
        # sorting shouldn't matter for a count query
        fields_sort = self.fields_sort
        self.fields_sort = self.fields_sort_class()

        # more than one selected field will cause the count query to error out
        fields_select = self.fields_select
        self.fields_select = self.fields_select_class()

        # setting bounds causes count(*) to return 0 in both Postgres and
        # SQLite
        bounds = self.bounds
        self.bounds = self.bounds_class()

        ret = await self.interface.count(self.schema, self, **kwargs)

        # restore previous values now that count is done
        self.fields_sort = fields_sort
        self.fields_select = fields_select
        self.bounds = bounds

        # now we are going to compensate for the bounds being set
        if self.bounds:
            offset = self.bounds.offset
            ret -= offset
            if self.bounds.has_limit():
                limit = self.bounds.limit
                if ret > limit:
                    ret = limit

        return ret

    async def has(self, **kwargs):
        """returns true if there is atleast one row in the db matching the
        query, False otherwise

        :returns: bool
        """
        v = await self.one(**kwargs)
        return True if v else False

    async def insert(self, **kwargs):
        """persist the .fields that were set with .set_field and .set

        :returns: int|str|None, the primary key of the inserted row if it
            exists
        """
        return await self.interface.insert(
            self.schema,
            self.fields_set.todict(),
            **kwargs
        )

    async def update(self, **kwargs):
        """persist the .fields set in .set and .set_field using .fields_where
        """
        return await self.interface.update(
            self.schema,
            self.fields_set.todict(),
            self,
            **kwargs
        )

    async def upsert(self, conflict_field_names=None, **kwargs):
        """persist the .fields set with .set and .set_field

        The insert fields are all the fields set in .fields.

        The update fields are all the fields set in .fields minus the conflict
        fields

        :param conflict_field_names: list, the list of field names that will be
            checked for conflicts and will insert if these field values don't
            already exist in the db or update if they do
        :param **kwargs: passed through to the interface upsert method
        :returns: str|int|None, the primary key of the inserted/updated row
        """
        insert_fields = self.fields_set.todict()
        update_fields = dict(insert_fields)

        if not conflict_field_names:
            conflict_field_names = self.schema.pk_names

        for field_name in conflict_field_names:
            update_fields.pop(field_name, None)

        return await self.interface.upsert(
            self.schema,
            insert_fields,
            update_fields,
            conflict_field_names,
            **kwargs
        )

    async def delete(self, **kwargs):
        """remove fields matching the where criteria"""
        return await self.interface.delete(
            self.schema,
            self,
            **kwargs
        )




# -*- coding: utf-8 -*-
import datetime
import functools
import logging
import uuid
import re

from testdata.base import TestData
from datatypes import (
    OrderedSubclasses,
    ReflectModule,
    ReflectClass,
)

from ..model import Orm
from ..interface import get_interfaces
from ..exception import UniqueError


logger = logging.getLogger(__name__)


class ModelData(TestData):
    """Provides testdata hooks for the projects models, testdata is a python
    module for generating random data for testing

    https://github.com/Jaymon/testdata

    In order for this to be used, it needs to be loaded into memory so testdata
    can discover it:

        from prom.extras.testdata import ModelData

    I usually import this in a common test module that gets imported by all the
    actual tests so this is always loaded.

    This uses a lot of magic and I'm not sure the best way to describe it,
    basically, this will create get_*, create_*, and get_*_fields methods for
    any Orm subclass you have loaded into memory. The * will correspond to
    Orm.model_name and Orm.models_name (get_*_fields only ever corresponds to
    Orm.model_name).

    NOTE -- Because this class only creates those methods for Orm classes
    loaded into memory you need to import the modules for the classes you want
    to test

    Maybe the best way to understand what this does is by example.

    :Example:
        from prom import Orm

        class Foobar(Orm):
            pass

        import testdata
        from prom.extras.testdata import ModelData

        o = await testdata.get_foobar()
        print(o.pk) # None because .save was not called

        o = await testdata.create_foobar()
        print(o.pk) # pk will be set because .save() was called

        fields = await testdata.get_foobar_fields()
        print(fields) # dict that can be used create a Foobar(fields) instance

        # will return a list of 2 Foobar instances
        os = await testdata.get_foobars(foobar_count=2)

        # will return a list of 2 created Foobar instances
        os = await testdata.create_foobars(foobar_count=2)


    To customize a method, you can extend this class and define the method. The
    signature for any overridden methods is (self, orm_class, **kwargs),
    and you will usually just want to override the get_<MODELS-NAME>_fields
    class since that is the one the get_* and create_* methods use to populate
    the orms

    All the signatures of all the methods need to be the same because they are
    mixed and matched, so if you have a `Foobar` Orm class and override the
    get_foobar_fields testdata method, then get_foobar() would call
    get_foobar_fields and get_orm(Foobar, **kwargs) would also call
    get_foobar_fields. So all of these methods can be mixed and matched

    :Example:
        from prom.extras.testdata import ModelData

        class MyModelData(ModelData):
            async def get_foobar_fields(self, orm_class, **kwargs):
                return await super().get_orm_fields(orm_class, **kwargs)

    If you'd like to customize the testdata method names, you can set the
    Orm.model_name or the Orm.models_name class properties

    :Example:
        from prom import Orm

        class Foobar(Orm):
            model_name = "foo_bar"
            models_name = "foo_bars"

    So this is the order of calls:

        * .create_orms
            * .get_orms
                * .get_orm
                    * .get_fields
                    * .create_orm_instance

        * .create_orm
            * .get_orm
                * .get_fields
                * .create_orm_instance

    Any kwargs you pass in any of the methods will be passed to the methods
    below, so if you pass in a value to .create_orm it will be available in
    .get_fields
    """
    def _orm_classes(self):
        """Iterate through a list of orms that should be injected into testdata

        by default, we want to ignore any orm classes that are defined in this
        library since they are by definition base classes

        :returns: generator[Orm], each orm class
        """
        Orm.orm_classes.insert_modules()

        for orm_class in Orm.orm_classes.edges():
            yield orm_class

    def _gets_count(self, orm_class, **kwargs):
        """Find how many orm_class should be created

        :param orm_class: Orm
        :param **kwargs: this will be checked for the correct *_count or count
            value
        :returns: int, how many of orm_class is wanted
        """
        return kwargs.get(
            f"{orm_class.model_name}_count",
            kwargs.get(
                f"{orm_class.models_name}_count",
                kwargs.get(
                    "count",
                    1
                )
            )
        )

    async def _dispatch_method(self, orm_class, method, **kwargs):
        """Internal dispatch method. This uses the orm_class to first try and
        call a magical wrapper method (eg, if orm_class was name Foo and method
        was .get_orm then this would first try and run .get_foo and only then
        would it fallback to method

        :param orm_class: type, the Orm child
        :param method: callable, the fallback method
        :param **kwargs: passed through to whatever method is ran
        :returns: Any, whatever the ran method returns
        """
        if isinstance(orm_class, str):
            orm_class = self.get_orm_class(orm_class)

        method_names = set(dir(self))
        m = re.match(r"^([^_]+)_(orms?)(?:_(.+))?$", method.__name__)

        for oc in Orm.orm_classes.getmro(orm_class):
            parts = [m.group(1)]

            if m.group(2) == "orm":
                parts.append(oc.model_name)

            else:
                parts.append(oc.models_name)

            if suffix := m.group(3):
                parts.append(suffix)

            method_name = "_".join(parts)
            if method_name in method_names:
                method = getattr(self, method_name)
                break

        logger.debug(
            "Running {} as {} for orm_class {}".format(
                method.__name__,
                method_name,
                orm_class.__name__,
            )
        )

        kwargs.setdefault("orm_class", orm_class)
        return await method(**kwargs)

    def _parse_dispatch(self, method_name):
        """Parses method name and returns the found orm_class and the
        default/fallback method that can be passed to ._dispatch_method

        NOTE -- this is where a lot of the parsing magic happens, for adding
        functionality in the future this is probably the method to start with

        :param method_name: str, the full method name that will be parsed
        :returns: tuple[type, callable], (orm_class, method), this will return
            (None, None) if method_name isn't valid
        """
        orm_class = method = None

        parts = method_name.split("_")

        if len(parts) > 1 and parts[1] not in set(["orm", "orms"]):
            prefix = parts[0]
            model_name = suffix = ""
            method_names = set(dir(self))

            method_name = "_".join([prefix, "orm", parts[-1]])
            if method_name in method_names:
                model_name = "_".join(parts[1:-1])
                suffix = parts[-1]

            if not model_name:
                method_name = "_".join([prefix, "orm"])
                if method_name in method_names:
                    model_name = "_".join(parts[1:])
                    suffix = ""

            if model_name:
                if orm_class := self.get_orm_class(model_name, None):
                    if orm_class.model_name == model_name:
                        parts = [prefix, "orm"]

                    else:
                        parts = [prefix, "orms"]

                    if suffix:
                        parts.append(suffix)

                    method = getattr(self, "_".join(parts))

        return orm_class, method

    def __getattribute__(self, method_name):
        """Introspect method_name to see if it is a valid Orm reflection
        request, if it isn't then pass the call on down the line

        To be even more magical this will try and parse method_name and if
        it is valid it will return a partial that has the orm_class property
        set to make it easier to call without having to worry about things like
        having the right orm_class

        :param method_name: str, the method name we're looking for, if this
            isn't actually a method_name then it will pass it on down the line
        :returns: Any, if it successfully identified method_name then this will
            return a callable, otherwise it's whatever on down the line returns
        """
        if method_name.startswith("_"):
            return super().__getattribute__(method_name)

        orm_class = default_method = method = None

        try:
            method = super().__getattribute__(method_name)

        except AttributeError:
            logger.debug(
                f"Finding {self.__class__.__name__}.{method_name} method"
            )

            orm_class, default_method = self._parse_dispatch(method_name)

            if default_method:
                logger.debug(
                    f"Found {default_method.__name__} for {orm_class.__name__}"
                )

            else:
                raise

        if callable(method) or default_method:
            if not orm_class:
                orm_class, _ = self._parse_dispatch(method_name)

            if orm_class:
                def wrapper(**kwargs):
                    # https://github.com/Jaymon/prom/issues/166
                    # we want to override the passed in orm_class if it
                    # doesn't match our found orm class because this has
                    # most likely been called internally by another magic
                    # method that just passed kwargs
                    kwargs.setdefault("orm_class", orm_class)

                    if not isinstance(orm_class, kwargs["orm_class"]):
                        kwargs["orm_class"] = orm_class

                    # if these are always there it's easier to mess with them
                    # in child methods
                    kwargs.setdefault("fields", {})
                    kwargs.setdefault("attributes", {})

                    if method:
                        return method(**kwargs)

                    else:
                        return self._dispatch_method(
                            method=default_method,
                            **kwargs
                        )

                wrapper.__name__ = f"wrapped_getattribute_{method_name}"
                return wrapper

        return method

    async def close_orm_interfaces(self):
        """Close down all the globally created interfaces

        This, along with the unsafe_* methods are designed to be used in actual
        project testing, this is most useful in the test's asyncTearDown method
        """
        for inter in get_interfaces().values():
            await inter.close()

    async def unsafe_delete_orm_tables(self):
        """This will delete all the tables from the db

        NOTE -- You'll want to make sure you only call the method in the
        right environments as this really will delete all the tables in
        whatever dbs it has connections for
        """
        for inter in get_interfaces().values():
            await inter.unsafe_delete_tables()

    async def unsafe_install_orms(self, modpaths=None):
        """Go through and install all the Orm subclasses found in the passed in
        module paths

        This exists because if you are making use of a lot of transactions
        that don't nest that can cause prom's built-in error handling to
        completely fail so it's easier to create all the tables in the 
        beginning than rely on prom to magically do it

        :param modpaths: Sequence[str], a list of modpaths (eg ["foo.bar",
            "che"])
        """
        # import the module paths to load the Orms into memory
        if modpaths:
            Orm.orm_classes.insert_modules(modpaths)

        # now go through all the orm classes that have been loaded and install
        # them
        seen_table_names = set()
        for orm_class in self._orm_classes():
            for s in orm_class.schema.schemas:
                if s.table_name not in seen_table_names:
                    await s.orm_class.install()
                    seen_table_names.add(s.table_name)

    async def unsafe_reset_orms(self, modpaths=None):
        """Delete all the tables in the db and then load all the Orm child
        classes and make sure they have a table

        NOTE -- this is incredibly unsafe, it deletes and creates tables in
        the configured db

        :param modpaths: passed through to .unsafe_install_orms
        """
        await self.unsafe_delete_orm_tables()
        await self.unsafe_install_orms(modpaths=modpaths)

    def assure_orm_field_names(self, orm_class, **kwargs):
        """Field instances can have aliases, in order to allow you to pass in
        aliases, this will go through kwargs and normalize the field names

        :Example:
            class Foobar(Orm):
                che = Field(str, aliases=["baz"]

            kwargs = testdata.assure_orm_field_names(Foobar, {"baz": "1"})
            print(kwargs["che"]) # "1"

        :param orm_class: Orm
        :param **kwargs: the fields where keys will be normalized to field
            names in orm_class.schema
        :returns: dict, the normalized kwargs
        """
        schema = orm_class.schema
        # normalize passed in field names to make sure we correctly find the
        # field's value if it exists
        for field_name in list(kwargs.keys()):
            if schema.has_field(field_name):
                kwargs.setdefault(
                    schema.field_name(field_name),
                    kwargs.pop(field_name)
                )

        return kwargs

    def assure_ref_field_names(self, orm_class, ref_class, **kwargs):
        """Make sure the kwargs destined for orm_class don't impact the kwargs
        that will be used to create a ref_class instance

        Sometimes, we have multiple Orm classes that have the same field names
        and when those fields get set for orm_class they can cause ref_class to
        fail so we want to strip those common field names out

        :param orm_class: Orm, the main orm class
        :param ref_class: Orm, the Orm that orm_class references in some way
        :param **kwargs: the fields that were passed in for orm_class, if these
            fields are also in ref_class they will be stripped
        :returns: dict, the kwargs suitable to be used to create an instance of
            ref_class
        """
        ref_kwargs = {}
        orm_fields = orm_class.schema.fields
        ref_fields = ref_class.schema.fields
        for field_name, field_value in kwargs.items():
            if (field_name not in orm_fields) or (field_name not in ref_fields):
                ref_kwargs[field_name] = field_value

        return ref_kwargs

    async def assure_orm_refs(self, orm_class, assure_orm_class=None, **kwargs):
        """When creating an orm, they will often need foreign key values, this
        will go through any of the foreign key ref fields and create a foreign
        key if it wasn't included.

        This is a recursive method, when it finds a ref_class it will assure
        that ref's foreign keys before handling ref, this way all references all
        the way down the stack can be generated for the top level so the same
        refs can propogate all the way down the dependency chain

        :param orm_class: Orm
        :param assure_orm_class: Orm|None, the orm that is getting its
            references assured. If this is None then orm_class is considered the
            orm that is being checked
        :param **kwargs: orm_class's actual field name value will be checked and
            the ref's orm_class.model_name will be checked
            * ignore_refs: bool, default False, if True then refs will not be
                checked, passed in refs will still be set
            * require_fields: bool, default True, if True then create missing
                refs, if False, then refs won't be created and so if they are
                missing their fields will not be populated
            * ignore_field_names: set[str]|list[str], a set of field names that
                should be ignored when creating refs
        :returns: dict, the kwargs with ref field_name and ref
            orm_class.model_name will be included
        """
        logger.debug(f"Assuring orm refs for orm_class {orm_class.__name__}")
        kwargs = self.assure_orm_field_names(orm_class, **kwargs)

        # if assure class isn't passed in then we assume the passed in orm_class
        # is the class to be assured and all recursive calls will now have it
        # set
        if assure_orm_class is None:
            assure_orm_class = orm_class

        ignore_refs = kwargs.get("ignore_refs", False)
        require_fields = kwargs.get("require_fields", True)
        ignore_field_names = set(kwargs.get("ignore_field_names", []))

        for field_name, field in orm_class.schema.fields.items():
            if ref_class := field.ref:
                ref_field_name = ref_class.model_name

                if field_name in kwargs:
                    if ref_field_name not in kwargs:
                        kwargs[ref_field_name] = await ref_class.query.eq_pk(
                            kwargs[field_name]
                        ).one()

                elif field_name in ignore_field_names:
                    # we were explicitely told to ignore this field
                    continue

                else:
                    if ref_field_name in kwargs:
                        kwargs[field_name] = kwargs[ref_field_name].pk

                    else:
                        if not ignore_refs:
                            if (
                                require_fields
                                or field.is_required()
                                or self.yes()
                            ):
                                # handle all ref_class's refs before we handle
                                # ref_class
                                kwargs.update(await self.assure_orm_refs(
                                    ref_class,
                                    assure_orm_class,
                                    **kwargs
                                ))

                                fields = await self._dispatch_method(
                                    ref_class,
                                    self.create_orm,
                                    **self.assure_ref_field_names(
                                        assure_orm_class,
                                        ref_class,
                                        **kwargs
                                    )
                                )

                                kwargs[ref_field_name] = fields
                                kwargs[field_name] = kwargs[ref_field_name].pk

        return kwargs

    async def create_orm(self, orm_class, **kwargs):
        """create an instance of the orm and save it into the db

        The typical method call hierarchy from here:

            <CREATE-ORM>
                <GET-ORM>
                [<CREATE-ORMS>]

        :param orm_class: Orm
        :param **kwargs:
        :returns: Orm, the orm saved into the db
        """
        kwargs.setdefault("ignore_refs", False)
        instance = await self._dispatch_method(
            orm_class,
            self.get_orm,
            **kwargs
        )
        try:
            await instance.save(nest=True)

        except UniqueError as e:
            logger.warning(" ".join([
                f"Creating {orm_class.__name__} failed because it exists.",
                "Fetching the existing instance without updating it",
            ]))
            instance = await instance.requery()

        for k in list(kwargs.keys()):
            # We want to create any orms with FK references to orm_class if
            # counts were passed in
            if k.endswith("_count"):
                k_model_name = k[0:-6]
                k_orm_class = self.get_orm_class(k_model_name)
                if k_orm_class:
                    logger.debug(
                        "Creating {} {} instances tied to {} instance".format(
                            kwargs[k],
                            k_orm_class.__name__,
                            orm_class.__name__,
                        )
                    )
                    kwargs.setdefault(instance.model_name, instance)
                    await self.create_orms(k_orm_class, **kwargs)

        return instance

    async def create_orms(self, orm_class, **kwargs):
        """create instances of the orm and save it into the db

        :param orm_class: Orm
        :param **kwargs:
            count: int, how many instances you want
            <MODEL_NAME>_count: int, alias for count
        :returns: list, a list of Orm instances that have all been saved into
            the db
        """
        kwargs.setdefault("ignore_refs", False)
        instances = await self._dispatch_method(
            orm_class,
            self.get_orms,
            **kwargs
        )
        for instance in instances:
            await instance.save(nest=True)
        return instances

    def get_orm_class(self, model_name, *default, **kwargs):
        """get the orm class found at model_name, if model_name doesn't exist
        than return default if it exists

        :param model_name: str, the name of the orm class
        :param *default: Any, if exists then this will be returned instead of
            raising a ValueError if model_name isn't found
        :returns: Orm, the orm_class.model_name that matches model_name
        """
        try:
            return Orm.orm_classes[model_name]

        except KeyError:
            if default:
                return default[0]

            else:
                raise

    async def get_orm(self, orm_class, **kwargs):
        """get an instance of the orm but don't save it into the db

        The method call hierarchy from here:

            <GET-ORM>
                <GET-ORM-FIELDS>
                <CREATE-ORM-INSTANCE>

        :param orm_class: Orm
        :param **kwargs:
        :returns: Orm
        """
        kwargs.setdefault("ignore_refs", True)
        instance = kwargs.get(orm_class.model_name, None)
        if not instance:
            kwargs["fields"] = await self._dispatch_method(
                orm_class,
                self.get_orm_fields,
                **kwargs
            )

            # since .get_orm_fields is the most common overrided method, this
            # is just a nice message to notify that you most likely forgot to
            # return something
            if kwargs["fields"] is None:
                raise ValueError(" ".join([
                    "Dispatched {}".format(orm_class.model_name),
                    "get fields method returned None instead of dict"
                ]))

            # If get fields method returns an attribute dict we want to add
            # it to the passed in attributes dict, this is just syntactic
            # sugar so child classes only really ever need to override the
            # fields method
            attributes = kwargs["fields"].get("attributes", {})
            attributes.update(kwargs.get("attributes", {}))
            kwargs["attributes"] = attributes

            if kwargs["fields"] and "**" in kwargs["fields"]:
                kwargs.update(kwargs["fields"].pop("**"))

            instance = await self._dispatch_method(
                orm_class,
                self.create_orm_instance,
                **kwargs
            )

        return instance

    async def get_orms(self, orm_class, **kwargs):
        """get instances of the orm

        :param orm_class: Orm
        :param **kwargs:
            * count: int, how many instances you want
            * <MODEL_NAME>_count: int, alias for count
            * related_refs: bool, default True, all the orms will have the same
                foreign key references
        :returns: list, a list of Orm instances
        """
        ret = []
        if kwargs.get("related_refs", True):
            # because we need related refs, we will need to create refs if they
            # don't exist
            kwargs.setdefault("ignore_refs", False)
            kwargs = await self.assure_orm_refs(orm_class, **kwargs)

        count = self._gets_count(orm_class, **kwargs)
        for _ in range(count):
            ret.append(await self._dispatch_method(
                orm_class,
                self.get_orm,
                **kwargs
            ))

        return ret

    async def create_orm_instance(self, orm_class, **kwargs):
        """Semi-internal method to actually create an instance of orm_class

        This is semi-internal because it can be overridden and customized but
        it shouldn't be called externally, it is designed to only be called 
        internally by the other methods but never to be called externally

        https://github.com/Jaymon/prom/issues/170

        :param orm_class: Orm
        :param fields: dict[str, Any], these should ruoughly correspond to the
            Orm's Schema fields 
        :param **kwargs:
            * attributes: dict[str, Any], These will be set onto the instance
                after it is created
            * fields: dict[str, Any], these are the actual fields that will be
                passed to Orm.__init__, that makes this kwargs different than
                other methods because the fields have been separated out.
                Basically, if you did .get_orm_class(FooOrm, bar="...") then
                by the time we got to here you would access bar with
                kwargs["fields"]["bar"] instead of just kwargs["bar"] in all
                the other methods
        :returns: Orm, the actual instance populated with fields 
        """
        instance = orm_class(kwargs["fields"])

        if attributes := kwargs.get("attributes", {}):
            for name, value in attributes.items():
                setattr(instance, name, value)

        return instance

    async def get_orm_fields(self, orm_class, **kwargs):
        """Get the fields of an orm_class

        The method call hierarchy from here:

            <GET-ORM-FIELDS>
                <ASSURE-ORM-REFS>
                <GET-ORM-FIELD-VALUE>

        :param orm_class: Orm
        :param **kwargs: the fields found in orm_class.schema
            * require_fields: bool, default True, this will require that all
                fields have values even if they aren't required, this does not
                apply to foreign key references
            * field_callbacks: dict, the key is the field name and the value is
                a callable that can take self
            * ignore_field_names: set[str]|list[str], a set of field names that
                should be ignored when creating refs
            * fields: dict[str, Any], these will be used to seed the return
                dict, you use this to get non schema fields to be passed to
                orm's __init__ method
        :returns: dict, these are the fields that will be passed to
            Orm.__init__ with one exception, if the dict contains a key "**"
            then that key will be popped and it's value (which should be a
            dict) will update the kwargs that are passed to
            .create_orm_instance to create the actual orm instance. The "**"
            key is just a handy way to do all customizing in one overridden
            method, certain returned keys do certain things:
                * **: dict, yes, that's right, a double asterisks key will
                    move any of the value in the dict value back into kwargs
                    that will be passed to the .create_orm_instance method
                * attributes: dict, these will be merged with the **kwargs
                    attributes (with kwargs taking precedence) and is a handy
                    way for this method to pass values that will be set on
                    the orm instance
        """
        ret = kwargs.pop("fields", {})
        schema = orm_class.schema

        kwargs.setdefault("ignore_refs", True)
        kwargs = await self.assure_orm_refs(orm_class, **kwargs)

        ignore_field_names = set(kwargs.get("ignore_field_names", []))

        for field_name, field in schema.fields.items():
            if field_name in kwargs:
                # this value was passed in so we don't need to do anything
                ret[field_name] = kwargs[field_name]

            elif field_name in ignore_field_names:
                # we were explicitely told to ignore this field
                pass

            elif field.is_auto():
                # db will handle any auto-generating fields
                pass

            elif field.is_pk():
                # primary key isn't auto-generating and wasn't passed in, so
                # we'll cross our fingers and hope it will be taken care of
                # somewhere else
                pass

            elif field.is_ref():
                # foreign keys are handled in .assure_orm_refs
                pass

            elif not self.get_orm_field_required(field_name, field, **kwargs):
                pass

            else:
                ret[field_name] = self.get_orm_field_value(
                    field_name,
                    field,
                    fields=ret,
                    **kwargs
                )

        return ret

    def get_orm_field_value(self, field_name, field, **kwargs):
        """Returns the generated value for the specific field

        This is a wrapper around all the specific field type generators, see
        the other get_orm_field_* methods

        :param field_name: str
        :param field: Field, the orm's field property
        :param **kwargs: see .get_orm_fields for values this can have
        :returns: Any, the generated value
        """
        if field.choices:
            ret = self.get_orm_field_choice(
                field_name,
                field,
                **kwargs
            )

        else:
            field_callbacks = kwargs.get("field_callbacks", {})
            field_type = field.interface_type

            cb = getattr(
                field,
                "testdata",
                field.options.get(
                    "testdata",
                    field_callbacks.get(field_name, None)
                )
            )
            if cb:
                ret = cb(self)

            else:
                if issubclass(field_type, bool):
                    ret = self.get_orm_field_bool(
                        field_name,
                        field,
                        **kwargs
                    )

                elif issubclass(field_type, int):
                    ret = self.get_orm_field_int(
                        field_name,
                        field,
                        **kwargs
                    )

                elif issubclass(field_type, str):
                    ret = self.get_orm_field_str(
                        field_name,
                        field,
                        **kwargs
                    )

                elif issubclass(field_type, dict):
                    ret = self.get_orm_field_dict(
                        field_name,
                        field,
                        **kwargs
                    )

                elif issubclass(field_type, float):
                    ret = self.get_orm_field_float(
                        field_name,
                        field,
                        **kwargs
                    )

                elif issubclass(field_type, datetime.datetime):
                    ret = self.get_orm_field_datetime(
                        field_name,
                        field,
                        **kwargs
                    )

                elif issubclass(field_type, datetime.date):
                    ret = self.get_orm_field_date(
                        field_name,
                        field,
                        **kwargs
                    )

                elif issubclass(field_type, uuid.UUID):
                    ret = self.get_orm_field_uuid(
                        field_name,
                        field,
                        **kwargs
                    )

                else:
                    ret = self.get_orm_field_any(
                        field_name,
                        field,
                        **kwargs
                    )

        return ret

    def get_orm_field_required(self, field_name, field, **kwargs):
        """Returns True if field is required"""
        require_fields = kwargs.get("require_fields", True)
        return require_fields or field.is_required() or self.yes()

    def get_orm_field_choice(self, field_name, field, **kwargs):
        """Returns one of field's defined choices"""
        return self.choice(field.choices)

    def get_orm_field_bool(self, field_name, field, **kwargs):
        return bool(self.yes())

    def get_orm_field_int(self, field_name, field, **kwargs):
        size_info = field.size_info()
        return self.get_posint(
            size_info["size"]
        )

    def get_orm_field_str(self, field_name, field, **kwargs):
        size_info = field.size_info()
        if "bounds" in size_info:
            ret = self.get_words(
                min_size=size_info["bounds"][0],
                max_size=size_info["bounds"][1],
            )

        else:
            ret = self.get_words()

        return ret

    def get_orm_field_dict(self, field_name, field, **kwargs):
        return self.get_dict()

    def get_orm_field_float(self, field_name, field, **kwargs):
        size_info = field.size_info()
        return self.get_posfloat(
            size_info["size"]
        )

    def get_orm_field_datetime(self, field_name, field, **kwargs):
        return self.get_past_datetime()

    def get_orm_field_date(self, field_name, field, **kwargs):
        return self.get_past_date()

    def get_orm_field_uuid(self, field_name, field, **kwargs):
        return str(uuid.uuid4())

    def get_orm_field_any(self, field_name, field, **kwargs):
        """If one of the other field type generators isn't called then this
        will be called. It's designed for children classes to customize the
        field value generator further and support more types"""
        raise ValueError(
            f"Not sure what to do with {field.type}"
        )




# -*- coding: utf-8 -*-
import os
import datetime
import decimal
import logging
import uuid
from functools import cached_property
import inspect

from ..query import Query, QueryField
from ..exception import (
    TableError,
    FieldError,
    UniqueError,
    PlaceholderError,
)

from ..compat import *
from ..utils import make_list
from .base import Connection, Interface


logger = logging.getLogger(__name__)


class SQLConnection(Connection):
    """
    https://docs.python.org/3.9/library/sqlite3.html#sqlite3-controlling-transactions
    https://www.sqlite.org/lockingv3.html
    https://www.sqlite.org/lang_transaction.html
    """
    async def _transaction_start(self):
        await self.execute("BEGIN")

    async def _transaction_starting(self, tx):
        # http://www.postgresql.org/docs/9.2/static/sql-savepoint.html
        await self.execute("SAVEPOINT {}".format(
            self.interface.render_field_name_sql(tx["name"])
        ))

    async def _transaction_stop(self):
        """
        http://initd.org/psycopg/docs/usage.html#transactions-control
        https://news.ycombinator.com/item?id=4269241
        """
        await self.execute("COMMIT")

    async def _transaction_stopping(self, tx):
        await self.execute("RELEASE {}".format(
            self.interface.render_field_name_sql(tx["name"])
        ))

    async def _transaction_fail(self):
        await self.execute("ROLLBACK")

    async def _transaction_failing(self, tx):
        # http://www.postgresql.org/docs/9.2/static/sql-rollback-to.html
        await self.execute("ROLLBACK TO SAVEPOINT {}".format(
            self.interface.render_field_name_sql(tx["name"])
        ))


class SQLInterfaceABC(Interface):
    """SQL database interfaces should extend SQLInterface and implement all
    these methods in this class and all the methods in InterfaceABC"""
    @property
    def LIMIT_NONE(self):
        """When an offset is set but not a limit, this is the value that will be
        put into the LIMIT part of the query

        :returns: str|int|None
        """
        raise NotImplementedError(
            "this property should be set in a child class"
        )

    def get_paramstyle(self):
        """Returns the paramstyle that is used by self.PLACEHOLDER to decide
        what val placeholder to use when building queries. This would also be
        the value to use when calling .raw().

        The dbapi 2.0 spec requires this to be set

        https://peps.python.org/pep-0249/#paramstyle

        :returns: str, something like "qmark" (for ? placeholders) or "pyformat"
            (for %s placeholders)
        """
        raise NotImplemented()

    async def render_date_field_sql(self, field_name, field_kwargs, symbol):
        raise NotImplemented()

    async def render_sort_field_sql(self, field_name, field_vals, sort_dir_str):
        """normalize the sort string

        return -- tuple -- field_sort_str, field_sort_args"""
        raise NotImplemented()

    async def render_nolimit_sql(limit, **kwargs):
        raise NotImplemented()

    async def render_datatype_datetime_sql(self, field_name, field, **kwargs):
        raise NotImplementedError()

    async def render_datatype_dict_sql(self, field_name, field, **kwargs):
        raise NotImplementedError()

    async def render_datatype_uuid_sql(self, field_name, field, **kwargs):
        raise NotImplementedError()


class SQLInterface(SQLInterfaceABC):
    """Generic base class for all SQL derived interfaces"""
    @cached_property
    def PLACEHOLDER(self):
        """What placeholder value this interface uses when building queries.

        This uses .get_paramstyle() to decide what placeholder value to use

        https://www.psycopg.org/docs/usage.html#passing-parameters-to-sql-queries

        NOTE -- It looks like both SQLite and Postgres support "named" so if
        you want to use .raw() queries that will work in both interface I would
        use named parameters

        :returns: str, usually something like "?" or "%s"
        """
        paramstyle = self.get_paramstyle()

        if paramstyle == "qmark":
            return "?"

        elif paramstyle == "format":
            return "%s"

        elif paramstyle == "pyformat":
            return "%s"

        elif paramstyle == "numeric":
            raise NotImplementedError("These are :1 :2 :3")

        elif paramstyle == "named":
            raise NotImplementedError("These are :name")

        else:
            raise NotImplementedError(f"Unknown paramstyle {paramstyle}")

    async def _set_table(self, schema, **kwargs):
        """
        http://sqlite.org/lang_createtable.html
        http://www.postgresql.org/docs/9.1/static/sql-createtable.html
        http://www.postgresql.org/docs/8.1/static/datatype.html
        http://pythonhosted.org/psycopg2/usage.html#adaptation-of-python-values-to-sql-types
        """
        query_str = [
            "CREATE TABLE IF NOT EXISTS {} (".format(
                self.render_table_name_sql(schema)
            )
        ]

        query_fields = []
        for field_name, field in schema.fields.items():
            query_fields.append("  {}".format(
                self.render_datatype_sql(field_name, field)
            ))
        query_str.append(",\n".join(query_fields))

        query_str.append(")")
        query_str = "\n".join(query_str)
        await self._raw(query_str, ignore_result=True, **kwargs)

    def _set_index(self, schema, name, field_names, **kwargs):
        """
        NOTE -- we set the index name using <table_name>_<name> format since
        indexes have to have a globally unique name in postgres

        * https://www.sqlite.org/lang_createindex.html
        * https://www.postgresql.org/docs/14/sql-createindex.html - "IF NOT
            EXISTS support was added around 9.5
        """
        query_str = 'CREATE {}INDEX IF NOT EXISTS {} ON {} ({})'.format(
            'UNIQUE ' if kwargs.get('unique', False) else '',
            self.render_field_name_sql(f"{schema}_{name}"),
            self.render_table_name_sql(schema),
            ', '.join(map(self.render_field_name_sql, field_names))
        )

        return self._raw(query_str, ignore_result=True, **kwargs)

    async def _insert(self, schema, fields, **kwargs):
        query_str, query_args = self.render_insert_sql(
            schema,
            fields,
            **kwargs,
        )

        r = await self._raw(query_str, *query_args, **kwargs)
        return r[0] if r else {}

    async def _inserts(self, schema, field_names, field_values, **kwargs):
        """Do a mutli insert

        The query this will run is roughly:

            INSERT INTO
                (<field_names>)
            VALUES
                (tuple for tuple in <field_values>)

        :param schema: Schema, the table
        :param field_names: list[str], the field names
        :param field_values: Iterable[Iterable[Any]], each row in the iterable
            is a tuple of values that correspond to field_names
        :param **kwargs: passed through
        """
        query_str = self.render_inserts_sql(schema, field_names, **kwargs)
        await self._raw(
            query_str,
            *field_values,
            ignore_result=True,
            execute_many=True,
            **kwargs
        )

    async def _update(self, schema, fields, query, **kwargs):
        query_str, query_args = self.render_update_sql(
            schema,
            fields,
            query=query,
            **kwargs,
        )

        return await self._raw(
            query_str,
            *query_args,
            **kwargs
        )

    async def _upsert(
        self,
        schema,
        insert_fields,
        update_fields,
        conflict_field_names,
        **kwargs
    ):
        """
        https://www.sqlite.org/lang_UPSERT.html
        """
        if not conflict_field_names:
            raise ValueError(f"Upsert is missing conflict fields for {schema}")

        for field_name in conflict_field_names:
            # conflict fields need to be in the insert fields
            if field_name not in insert_fields:
                errmsg = "Upsert insert fields on {} missing conflict field {}"
                raise ValueError(
                    errmsg.format(
                        schema,
                        field_name,
                    )
                )

            # conflict fields should not be in the update fields (this is more
            # for safety, they should use .update if they want to change them)
            if field_name in update_fields:
                errmsg = "Upsert update fields on {} contain conflict field {}"
                raise ValueError(
                    errmsg.format(
                        schema,
                        field_name,
                    )
                )

        insert_sql, insert_args = self.render_insert_sql(
            schema,
            insert_fields,
            ignore_return_clause=True,
            **kwargs,
        )

        update_sql, update_args = self.render_update_sql(
            schema,
            update_fields,
            query=None,
            only_set_clause=True,
            **kwargs,
        )

        query_str = '{} ON CONFLICT({}) DO UPDATE {}'.format(
            insert_sql,
            ', '.join(map(self.render_field_name_sql, conflict_field_names)),
            update_sql,
        )

        returning_field_names = schema.pk_names
        if returning_field_names:
            # https://www.sqlite.org/lang_returning.html
            query_str += ' RETURNING {}'.format(', '.join(
                map(self.render_field_name_sql, returning_field_names))
            )
            query_args = insert_args + update_args

        r = await self._raw(query_str, *query_args, **kwargs)
        if r and returning_field_names:
            if len(returning_field_names) > 1:
                r = r[0]
            else:
                r = r[0][returning_field_names[0]]
        return r

    async def _delete(self, schema, query, **kwargs):
        where_query_str, query_args = self.render_sql(
            schema,
            query,
            only_where_clause=True
        )
        query_str = []
        query_str.append('DELETE FROM')
        query_str.append('  {}'.format(self.render_table_name_sql(schema)))
        query_str.append(where_query_str)
        query_str = "\n".join(query_str)
        return await self._raw(
            query_str,
            *query_args,
            count_result=True,
            **kwargs
        )

    async def _raw(self, query_str, *query_args, **kwargs):
        """
        :param **kwargs:
            - ignore_result: bool, true to not attempt to fetch results
            - fetchone: bool, true to only fetch one result
            - count_result: bool, true to return the int count of rows affected
            - execute_many: bool, True to call cursor.executemany instead of
                cursor.execute
        """
        ret = True
        # http://stackoverflow.com/questions/6739355/dictcursor-doesnt-seem-to-work-under-psycopg2
        #connection = kwargs.get('connection', None)
        async with self.connection(**kwargs) as connection:
            cur = connection.cursor()
            # depending on the api, cursor() could either return a coroutine
            # or something different like a cursor class instance
            if inspect.isawaitable(cur):
                cur = await cur

            ignore_result = kwargs.get("ignore_result", False)
            count_result = kwargs.get("count_result", False)
            one_result = kwargs.get(
                "fetchone",
                kwargs.get("one_result", False)
            )
            cursor_result = kwargs.get("cursor_result", False)
            execute_many = kwargs.get(
                "execute_many",
                kwargs.get("executemany", False)
            )

            if cursor_result:
                # https://stackoverflow.com/a/125140
                cur.arraysize = kwargs.get("arraysize", 500)

            if query_args:
                self.log_for(
                    debug=(
                        ["{} - {}\n{}", connection, query_str, query_args],
                    ),
                    info=(["{} - {}", connection, query_str],)
                )

                execute_args = [query_str, query_args]

            else:
                self.log_info("{} - {}", connection, query_str)

                execute_args = [query_str]

            try:
                if execute_many:
                    # https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.executemany
                    # https://www.psycopg.org/psycopg3/docs/api/cursors.html#psycopg.AsyncCursor.executemany
                    # https://www.psycopg.org/docs/cursor.html#cursor.executemany
                    await cur.executemany(*execute_args)

                else:
                    await cur.execute(*execute_args)

            except Exception as e:
                await self.raise_error(
                    e,
                    error_args=[
                        query_str,
                        query_args,
                        self.PLACEHOLDER,
                    ]
                )

            if cursor_result:
                ret = cur

            elif ignore_result:
                await cur.close()

            else:
                if one_result:
                    # https://www.psycopg.org/docs/cursor.html#cursor.fetchone
                    ret = await cur.fetchone()

                elif count_result:
                    # https://www.psycopg.org/docs/cursor.html#cursor.rowcount
                    ret = cur.rowcount

                else:
                    # https://www.psycopg.org/docs/cursor.html#cursor.fetchall
                    ret = await cur.fetchall()

                await cur.close()

            return ret

    async def _get(self, schema, query, **kwargs):
        """
        https://www.sqlite.org/lang_select.html
        """
        query_str, query_args = self.render_sql(schema, query)
        return await self._raw(query_str, *query_args, **kwargs)

    async def _count(self, schema, query, **kwargs):
        query_str, query_args = self.render_sql(
            schema,
            query,
            count_query=True
        )
        ret = await self._raw(query_str, *query_args, **kwargs)
        if ret:
            ret = int(ret[0]['ct'])
        else:
            ret = 0

        return ret

    async def _handle_field_error(self, schema, e, **kwargs):
        """This will add fields that don't exist in the table if they can be
        set to NULL, the reason they have to be NULL is adding fields to
        Postgres that can be NULL is really light, but if they have a default
        value, then it can be costly
        """
        current_fields = await self._get_fields(schema, **kwargs)
        for field_name, field in schema.fields.items():
            if field_name not in current_fields:
                if field.required:
                    self.log_error(
                        "Required field {} cannot be safely add on the fly",
                        field_name
                    )
                    return False

                else:
                    query_str = []
                    query_str.append('ALTER TABLE')
                    query_str.append('  {}'.format(
                        self.render_table_name_sql(schema)
                    ))
                    query_str.append('ADD COLUMN')
                    query_str.append('  {}'.format(
                        self.render_datatype_sql(field_name, field)
                    ))
                    query_str = "\n".join(query_str)
                    await self._raw(query_str, ignore_result=True, **kwargs)

        return True

    async def _handle_table_error(self, schema, e, **kwargs):
        """
        You can run into a problem when you are trying to set a table and it
        has a foreign key to a table that doesn't exist, so this method will go
        through all fk refs and make sure all the tables exist
        """
        if query := kwargs.pop("query", None):
            if schemas := query.schemas:
                for s in schemas:
                    self.log_warning(
                        f"Verifying {schema} query foreign key table: {s}"
                    )
                    if not await self._handle_table_error(s, e=e, **kwargs):
                        return False

        for field_name, field_val in schema.fields.items():
            s = field_val.schema
            if s:
                self.log_warning(f"Verifying {schema} foreign key table: {s}")
                if not await self._handle_table_error(schema=s, e=e, **kwargs):
                    return False

        # now that we know all fk tables exist, create this table
        # !!! This uses the external .set_table so it will run through all the 
        # indexes also
        await self.set_table(schema, **kwargs)
        return True

    def render_table_name_sql(self, schema):
        return self.render_field_name_sql(schema)

    def render_field_name_sql(self, name):
        """normalize a non value name for the query

        https://blog.christosoft.de/2012/10/sqlite-escaping-table-acolumn-names/

        :param name: str|QueryField|Schema, the name that should be prepared to
            be queried
        :returns: the modified name ready to be added to a query string
        """
        if isinstance(name, QueryField):
            if name.function_name:
                if name.alias:
                    return "{}(\"{}\") AS \"{}\"".format(
                        name.function_name,
                        name.name,
                        name.alias
                    )

                else:
                    return "{}(\"{}\")".format(
                        name.function_name,
                        name.name,
                    )

            else:
                return "\"{}\"".format(name.name)

        else:
            return "\"{}\"".format(name)

    def render_select_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        only_where_clause = kwargs.get('only_where_clause', False)
        if not only_where_clause:
            query_str.append('SELECT')
            is_count_query = kwargs.get('count_query', False)

            if query.compounds:
                select_fields_str = "*"

            else:
                select_fields = query.fields_select
                if select_fields:
                    distinct_fields = select_fields.options.get(
                        "distinct",
                        False
                    )
                    distinct = "DISTINCT " if distinct_fields else ""
                    select_fields_str = distinct + ", ".join(
                        (self.render_field_name_sql(f) for f in select_fields)
                    )

                else:
                    if is_count_query or select_fields.options.get("all", False):
                        select_fields_str = "*"

                    else:
                        select_fields_str = ", ".join(
                            (
                                self.render_field_name_sql(fname)
                                for fname in schema.fields.keys()
                            )
                        )

            if is_count_query:
                query_str.append('  count({}) as ct'.format(select_fields_str))

            else:
                query_str.append('  {}'.format(select_fields_str))

            query_str.append('FROM')

            if not query.compounds:
                query_str.append("  {}".format(
                    self.render_table_name_sql(schema)
                ))

        return query_str, query_args

    def render_compound_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        if query.compounds:
            query_str.append("(")
            for operator, queries in query.compounds:
                for i, query in enumerate(queries):
                    if i > 0:
                        query_str.append(operator.upper())

                    subquery_str, subquery_args = self.render_sql(
                        query.schema,
                        query,
                    )

                    if query.bounds:
                        query_str.append("  (")
                        query_str.append(subquery_str)
                        query_str.append("  )")

                    else:
                        query_str.append(subquery_str)

                    query_args.extend(subquery_args)

            query_str.append(") I")
        return query_str, query_args

    def render_where_field_raw_sql(self, schema, field):
        return field.name, field.value

    def render_where_field_sql(self, schema, field):
        format_str = ''
        format_args = []
        is_list = field.is_list
        field_name = field.name
        field_val = field.value

        symbol_map = {
            'in': {'symbol': 'IN', 'list': True},
            'nin': {'symbol': 'NOT IN', 'list': True},
            # Previously we used "IS" and "IS NOT" but psycopg3 changed behavior
            # https://stackoverflow.com/a/76396765
            # https://www.psycopg.org/psycopg3/docs/basic/from_pg2.html#you-cannot-use-is-s
            'eq': {'symbol': '=', 'none_symbol': 'IS NOT DISTINCT FROM'},
            'ne': {'symbol': '!=', 'none_symbol': 'IS DISTINCT FROM'},
            'gt': {'symbol': '>'},
            'gte': {'symbol': '>='},
            'lt': {'symbol': '<'},
            'lte': {'symbol': '<='},
            # https://www.tutorialspoint.com/postgresql/postgresql_like_clause.htm
            # https://www.tutorialspoint.com/sqlite/sqlite_like_clause.htm
            'like': {'symbol': 'LIKE'},
            'nlike': {'symbol': 'NOT LIKE'},
        }
        sd = symbol_map[field.operator]
        symbol = sd['symbol']

        field_kwargs = field.kwargs
        if field_kwargs:
            # kwargs take precedence because None is a perfectly valid field_val
            f = schema.fields[field_name]
            if issubclass(f.type, (datetime.datetime, datetime.date)):
                format_strs = self.render_date_field_sql(
                    field_name,
                    field_kwargs,
                    symbol
                )
                for fname, fvstr, farg in format_strs:
                    if format_str:
                        format_str += ' AND '

                    if is_list:
                        # you can pass in things like day=..., month=... to
                        # date fields, this converts those values to lists to
                        # make sure we can handle something like in_foo(day=1)
                        # and in_foo(day=[1, 2, 3]) the same way
                        farg = make_list(farg)

                        format_str += '{} {} ({})'.format(
                            fname,
                            symbol,
                            ', '.join([fvstr] * len(farg))
                        )
                        format_args.extend(farg)

                    else:
                        format_str += '{} {} {}'.format(fname, symbol, fvstr)
                        format_args.append(farg)

            else:
                raise ValueError(
                    'Field {} does not support extended kwarg values'.format(
                        field_name
                    )
                )

        else:
            if is_list and not isinstance(field_val, Query):
                field_val = make_list(field_val) if field_val else []
                field_name = self.render_field_name_sql(field_name)
                format_val_str = self.PLACEHOLDER

                if field_val:
                    format_str = '{} {} ({})'.format(
                        field_name,
                        symbol,
                        ', '.join([format_val_str] * len(field_val))
                    )
                    format_args.extend(field_val)

                else:
                    # field value is empty, so we need to customize the SQL to
                    # compensate for the empty set since SQL doesn't like empty
                    # sets
                    #
                    # the idea here is this is a condition that will
                    # automatically cause the query to fail but not necessarily 
                    # be an error, the best example is the IN (...) queries, if
                    # you do self.in_foo([]).get() that will fail because the
                    # list was empty, but a value error shouldn't be raised
                    # because a common case is:
                    #   self.if_foo(Bar.query.is_che(True).pks).get()
                    # which should result in an empty set if there are no rows
                    # where che = TRUE
                    #
                    # https://stackoverflow.com/a/58078468/5006
                    if symbol == "IN":
                        format_str = '{} <> {}'.format(field_name, field_name)

                    elif symbol == "NOT IN":
                        format_str = '{} = {}'.format(field_name, field_name)

                    else:
                        raise ValueError("Unsure what to do here")

            else:
                field_name = self.render_field_name_sql(field_name)
                format_val_str = self.PLACEHOLDER

                # special handling for NULL
                if field_val is None:
                    symbol = sd['none_symbol']

                if isinstance(field_val, Query):
                    subquery_schema = field_val.schema
                    if not subquery_schema:
                        raise ValueError(
                            f"{field_name} subquery has no schema"
                        )

                    subquery_sql, subquery_args = self.render_sql(
                        subquery_schema,
                        field_val
                    )

                    format_str = '{} {} ({})'.format(
                        field_name,
                        symbol,
                        subquery_sql
                    )
                    format_args.extend(subquery_args)

                else:
                    format_str = '{} {} {}'.format(
                        field_name,
                        symbol,
                        format_val_str,
                    )
                    format_args.append(field_val)

        return format_str, format_args

    def render_where_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        if query.fields_where:
            query_str.append('WHERE')
            or_clause = False

            for i, field in enumerate(query.fields_where):
                if i > 0:
                    query_str.append('OR' if or_clause else 'AND')

                field_str = ''
                field_args = []

                if field.raw:
                    field_str, field_args = self.render_where_field_raw_sql(
                        schema,
                        field,
                    )

                else:
                    field_str, field_args = self.render_where_field_sql(
                        schema,
                        field,
                    )

                if field.or_clause:
                    if not or_clause:
                        query_str.append("(")
                        or_clause = True

                query_str.append('  {}'.format(field_str))
                query_args.extend(field_args)

                if or_clause:
                    if not field.or_clause:
                        query_str.append(")")
                        or_clause = False

        return query_str, query_args

    def render_sort_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        if query.fields_sort:
            query_str.append('ORDER BY')

            query_sort_str = []
            for field in query.fields_sort:
                sort_dir_str = 'ASC' if field.direction > 0 else 'DESC'
                if field.value:
                    field_sort_str, field_sort_args = self.render_sort_field_sql(
                        field.name,
                        field.value,
                        sort_dir_str
                    )
                    query_sort_str.append(field_sort_str)
                    query_args.extend(field_sort_args)

                else:
                    query_sort_str.append('  {} {}'.format(
                        field.name,
                        sort_dir_str
                    ))

            query_str.append(',\n'.join(query_sort_str))

        return query_str, query_args

    def render_bounds_sql(self, schema, query, **kwargs):
        """
        https://www.postgresql.org/docs/current/queries-limit.html
        https://www.sqlite.org/lang_select.html#the_limit_clause
        """
        query_str = []
        query_args = []

        bounds = query.bounds
        fetchone = kwargs.get("fetchone", False)
        if bounds or fetchone:
            if fetchone:
                limit = 1
                offset = bounds.offset

            else:
                if bounds.has_limit():
                    limit, offset = bounds.get()

                else:
                    limit = self.LIMIT_NONE
                    offset = bounds.offset

            query_str.append(f'LIMIT {limit} OFFSET {offset}')

        return query_str, query_args

    def render_sql(self, schema, query, **kwargs):
        """
        convert the query instance into SQL

        this is the glue method that translates the generic Query() instance to
        the SQL specific query, this is where the magic happens

        https://www.sqlite.org/lang_select.html

        :param **kwargs:
            - count_query: bool, True if this is a count query SELECT
            - only_where_clause, bool, True to only return after WHERE ...
        :returns: tuple[str, list[Any]], (query_str, query_args)
        """
        query_str = []
        query_args = []

        select_str, select_args = self.render_select_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(select_str)
        query_args.extend(select_args)

        compound_str, compound_args = self.render_compound_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(compound_str)
        query_args.extend(compound_args)

        where_str, where_args = self.render_where_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(where_str)
        query_args.extend(where_args)

        sort_str, sort_args = self.render_sort_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(sort_str)
        query_args.extend(sort_args)

        limit_str, limit_args = self.render_bounds_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(limit_str)
        query_args.extend(limit_args)

        query_str = "\n".join(query_str)
        return query_str, query_args

    def render_subquery_sql(self, subquery):
        subquery_schema = subquery.schema
        if not subquery_schema:
            raise ValueError(f"Subquery has no schema")

        return self.render_sql(subquery_schema, subquery)

    def render_set_value_sql(self, field_val):
        if isinstance(field_val, QueryField):
            if field_val.increment:
                if field_val.value:
                    query_str, query_vals = self.render_set_value_sql(
                        field_val.value
                    )

                    # We use COALESCE here to make sure there is a value to
                    # start with if there isn't a starting value
                    # Given a list of values, the COALESCE function returns
                    # the first non-null value
                    # https://www.postgresql.org/docs/current/functions-conditional.html#FUNCTIONS-COALESCE-NVL-IFNULL
                    query_str = "COALESCE({}, 0) + {}".format(
                        query_str,
                        self.PLACEHOLDER
                    )
                    query_vals.append(field_val.increment)

                else:
                    query_str = "{} + {}".format(
                        self.render_field_name_sql(field_val.name),
                        self.PLACEHOLDER
                    )
                    query_vals = [field_val.increment]

            else:
                query_str, query_vals = self.render_set_value_sql(
                    field_val.value
                )

        elif isinstance(field_val, Query):
            subquery_sql, query_vals = self.render_subquery_sql(field_val)
            query_str = "({})".format(subquery_sql)

        else:
            query_str = self.PLACEHOLDER
            query_vals = [field_val]

        return query_str, query_vals

    def render_insert_sql(self, schema, fields, **kwargs):
        """
        https://www.sqlite.org/lang_insert.html
        """
        field_names = []
        field_values = []
        query_vals = []

        ignore_result = kwargs.get("ignore_result", False)
        ignore_return_clause = kwargs.get(
            "ignore_return_clause",
            ignore_result
        )

        for field_name, field_val in fields.items():
            field_names.append(self.render_field_name_sql(field_name))

            field_value, field_query_vals = self.render_set_value_sql(
                field_val
            )
            field_values.append(field_value)
            query_vals.extend(field_query_vals)

        query_str = "INSERT INTO {} ({}) VALUES ({})".format(
            self.render_table_name_sql(schema),
            ", ".join(field_names),
            ", ".join(field_values),
        )

        if not ignore_return_clause:
            # https://www.sqlite.org/lang_returning.html
            if pk_name := schema.pk_name:
                field_names.append(self.render_field_name_sql(pk_name))

            query_str += " RETURNING {}".format(
                ", ".join(field_names)
            )

        return query_str, query_vals

    def render_inserts_sql(self, schema, field_names, **kwargs):
        """
        https://www.sqlite.org/lang_insert.html
        """
        sql_names = []
        sql_formats = []
        for field_name in field_names:
            sql_names.append(self.render_field_name_sql(field_name))
            sql_formats.append(self.PLACEHOLDER)

        return "INSERT INTO {} ({}) VALUES ({})".format(
            self.render_table_name_sql(schema),
            ", ".join(sql_names),
            ", ".join(sql_formats),
        )

    def render_update_sql(self, schema, fields, query, **kwargs):
        """
        https://www.sqlite.org/lang_update.html
        https://www.postgresql.org/docs/current/sql-update.html
        """
        query_str = ''
        query_args = []
        returning_names = []

        count_result = kwargs.get("count_result", False)
        ignore_result = kwargs.get("ignore_result", False)
        only_set_clause = kwargs.get("only_set_clause", False)
        ignore_return_clause = kwargs.get(
            "ignore_return_clause",
            only_set_clause or ignore_result or count_result
        )

        if not only_set_clause:
            query_str = "UPDATE {} ".format(self.render_table_name_sql(schema))

        field_str = []
        for field_name, field_val in fields.items():
            field_name_query_str = self.render_field_name_sql(field_name)
            field_query_str, field_query_vals = self.render_set_value_sql(
                field_val
            )
            field_str.append("{} = {}".format(
                field_name_query_str,
                field_query_str,
            ))
            query_args.extend(field_query_vals)
            returning_names.append(field_name_query_str)

        query_str += "SET {}".format(",\n".join(field_str))

        if query:
            where_query_str, where_query_args = self.render_sql(
                schema,
                query,
                only_where_clause=True
            )
            query_str += " {}".format(where_query_str)
            query_args.extend(where_query_args)

        if not ignore_return_clause:
            if returning_names:
                query_str += " RETURNING {}".format(
                    ",\n".join(returning_names)
                )

        return query_str, query_args

    def render_datatype_sql(self, field_name, field):
        """Returns the SQL for a given field with full type information

        http://www.sqlite.org/datatype3.html
        https://www.postgresql.org/docs/current/datatype.html

        :param field_name: str, the field's name
        :param field: Field instance, the configuration for the field
        :returns: str, the complete field datatype SQL (eg, foo BOOL NOT NULL)
        """
        field_type = ""
        interface_type = field.interface_type

        if issubclass(interface_type, bool):
            field_type = self.render_datatype_bool_sql(field_name, field)

        elif issubclass(interface_type, int):
            field_type = self.render_datatype_int_sql(field_name, field)

        elif issubclass(interface_type, str):
            field_type = self.render_datatype_str_sql(field_name, field)

        elif issubclass(interface_type, datetime.datetime):
            field_type = self.render_datatype_datetime_sql(field_name, field)

        elif issubclass(interface_type, datetime.date):
            field_type = self.render_datatype_date_sql(field_name, field)

        elif issubclass(interface_type, dict):
            field_type = self.render_datatype_dict_sql(field_name, field)

        elif issubclass(interface_type, (float, decimal.Decimal)):
            field_type = self.render_datatype_float_sql(field_name, field)

        elif issubclass(interface_type, (bytearray, bytes)):
            field_type = self.render_datatype_bytes_sql(field_name, field)

        elif issubclass(interface_type, uuid.UUID):
            field_type = self.render_datatype_uuid_sql(field_name, field)

        else:
            raise ValueError('Unknown python type: {} for field: {}'.format(
                interface_type.__name__,
                field_name,
            ))

        field_type += ' ' + self.render_datatype_required_sql(field_name, field)

        if field.is_pk():
            field_type += ' PRIMARY KEY'

        else:
            if field.is_ref():
                field_type += ' ' + self.render_datatype_ref_sql(
                    field_name,
                    field
                )

        return '{} {}'.format(
            self.render_field_name_sql(field_name),
            field_type
        )

    def render_datatype_bool_sql(self, field_name, field, **kwargs):
        return 'BOOL'

    def render_datatype_int_sql(self, field_name, field, **kwargs):
        return 'INTEGER'

    def render_datatype_str_sql(self, field_name, field, **kwargs):
        fo = field.interface_options
        field_type = kwargs.get("datatype", 'TEXT')
        size_info = field.size_info()

        # https://www.sqlitetutorial.net/sqlite-check-constraint/
        if 'size' in size_info["original"]:
            field_type += f" CHECK(length({field_name}) = {size_info['size']})"

        elif 'max_size' in size_info["original"]:
            if "min_size" in size_info["original"]:
                field_type += f" CHECK(length({field_name}) >= {size_info['original']['min_size']}"
                field_type += " AND "
                field_type += f"length({field_name}) <= {size_info['original']['max_size']})"

            else:
                field_type += f" CHECK(length({field_name}) <= {size_info['size']})"

        return field_type

    def render_datatype_date_sql(self, field_name, field):
        return 'DATE'

    def render_datatype_float_sql(self, field_name, field, **kwargs):
        return 'REAL'

    def render_datatype_bytes_sql(self, field_name, field, **kwargs):
        return 'BLOB'

    def render_datatype_required_sql(self, field_name, field, **kwargs):
        return 'NOT NULL' if field.required else 'NULL'

    def render_datatype_ref_sql(self, field_name, field, **kwargs):
        ref_s = field.schema
        if field.required: # strong ref, it deletes on fk row removal
            format_str = 'REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE CASCADE'

        else: # weak ref, it sets column to null on fk row removal
            format_str = 'REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE SET NULL'

        ret = format_str.format(
            self.render_table_name_sql(ref_s),
            self.render_field_name_sql(ref_s.pk.name)
        )

        return ret

    def render(self, schema, query, **kwargs):
        """Render the query

        :param schema: Schema, the query schema
        :param query: Query, the query to render
        :param **kwargs: named arguments
            placeholders: boolean, True if place holders should remain
        :returns: str if placeholders is False, tuple[str, list] if
            placeholders is True
        """
        sql, sql_args = self.render_sql(schema, query)
        placeholders = kwargs.get(
            "placeholders",
            kwargs.get("placeholder", False)
        )

        if not placeholders:
            for sql_arg in sql_args:
                if sql_arg is None:
                    sa = 'NULL'

                else:
                    sa = String(sql_arg)
                    if not sa.isnumeric():
                        sa = "'{}'".format(sa)

                sql = sql.replace(self.PLACEHOLDER, sa, 1)

        return (sql, sql_args) if placeholders else sql




# -*- coding: utf-8 -*-
import importlib

from .compat import *


def get_objects(classpath, calling_classpath=""):
    """
    given a classpath like foo.bar.Baz return module foo.bar and class Baz
    objects

    .. seealso::
        https://docs.python.org/2.5/whatsnew/pep-328.html
        https://www.python.org/dev/peps/pep-0328/

    :param classpath: string, the full python class path (includes modules), a
        classpath is something like foo.bar.Che where Che is the class definied
        in the foo.bar module
    :param calling_classpath: string, if classpath is relative (eg, ..foo.Bar)
        then this is needed to resolve the relative classpath, it is usually the
        path of the module that is calling get_objects()
    :returns: tuple, (module, class)
    """
    if ":" in classpath:
        module_name, class_name = classpath.rsplit(":", 1)

    else:
        module_name, class_name = classpath.rsplit(".", 1)

    module = importlib.import_module(module_name, calling_classpath)
    try:
        klass = getattr(module, class_name)

    except AttributeError:
        raise AttributeError("module {} has no attribute {} parsing {}".format(
            module.__name__,
            class_name,
            classpath
        ))

    return module, klass


def make_list(vals):
    """make vals a list, no matter what

    :param vals: list, vals should always be a list or tuple (eg, *args passed
        as args)
    """
    ret = []
    if isinstance(vals, basestring):
        ret.append(vals)

    else:
        try:
            for val in vals:
                if isinstance(val, basestring):
                    ret.append(val)

                elif isinstance(val, (list, tuple)):
                    ret.extend(val)

                else:
                    try:
                        r = list(val)
                        ret.extend(r)

                    except TypeError:
                        ret.append(val)

        except TypeError:
            # TypeError: * is not iterable
            ret.append(vals)

    return ret


def make_dict(*fields):
    """lot's of methods take a dict or kwargs, this combines those

    Basically, we do a lot of def method(fields, **kwargs) and we want to merge
    those into one super dict with kwargs taking precedence, this does that

    fields -- dict -- a passed in dict
    fields_kwargs -- dict -- usually a **kwargs dict from another function

    return -- dict -- a merged fields and fields_kwargs
    """
    ret = {}

    for d in fields:
        if d:
            ret.update(d)

    return ret




# -*- coding: utf-8 -*-
from contextlib import asynccontextmanager
import uuid

from datatypes import (
    LogMixin,
    Stack,
)

from ..compat import *
from ..query import Query
from ..exception import (
    InterfaceError,
    UniqueError,
    TableError,
    FieldError,
    CloseError,
)


class ConnectionABC(LogMixin):
    """Subclasses should extend Connection and implement the methods in this
    class
    """
    async def _transaction_start(self):
        """Called when the first transaction is started"""
        pass

    async def _transaction_starting(self, tx):
        """Called when a nested transaction is started"""
        pass

    async def _transaction_ignoring(self, tx):
        """Called when nested transaction is ignored instead of started"""
        pass

    async def _transaction_stop(self):
        """Called when the last transaction is stopped"""
        pass

    async def _transaction_stopping(self, tx):
        """Called when a nested transaction is stopped"""
        pass

    async def _transaction_fail(self):
        """Called when the last transaction is failed"""
        pass

    async def _transaction_failing(self, tx):
        """Called when a nested transaction is failed"""
        pass


class Connection(ConnectionABC):
    """holds common methods that all raw connections should have"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # counting-ish semaphore, length will be greater than 0 if in a
        # transaction, and 0 if there are no current transactions.
        #
        # This will push every time transaction_start() is called, and
        # pop every time transaction_stop() is called.
        #
        # transaction_fail will clear this and rollback the transaction
        #
        # Holds the active transactions
        self.transactions = Stack()

    def transaction_count(self):
        """How many active transactions there currently are"""
        return len(self.transactions)

    def transaction_names(self, tx=None):
        """Get all the transaction names for logging

        :param tx: dict, sometimes you pop the transaction before you get the
            names, passing that popped tx will allow it to be placed at the
            end
        :returns: str, all the names nested
        """
        names = " > ".join((r["name"] for r in reversed(self.transactions)))

        if tx:
            if names:
                names += " > "

            names += tx["name"]

        return names

    def transaction_name(self, **kwargs):
        """generate a random transaction name for use in start_transaction() and
        fail_transaction()

        :param prefix: str, to better track transactions in logs you can give a
            prefix name that will be prepended to the auto-generated name
        """
        if not (name := kwargs.get("name", "")):
            suffix = str(uuid.uuid4())[-5:]
            prefix = kwargs.get("prefix", "") or "p"
            name = f"{prefix}_{suffix}"
        return name

    def transaction_info(self, **kwargs):
        """Create a new transaction dict that will be placed on the
        .transactions stack

        :param **kwargs:
            - nest: bool, True if this (and children unless passed in) will run
                nested transactions. If this is False then subsequent calls to
                .transaction_start will be ignored unless nest=True is passed in
                again
            - name: str, the transaction name
            - prefix: str, used to create a transaction name (see
              .transaction_name)
        :returns: dict[str], the created transaction with keys:
            - nest: bool, the value of kwargs["nest"] or of
              .transaction_current()["nest"]
            - name: str, the tx name
            - ignored: bool, True if this tx is going to be ignored
            - index: int, the depth of the transaction
        """
        name = self.transaction_name(**kwargs)
        current_tx = self.transaction_current()
        nest = kwargs.get("nest", current_tx.get("nest", True))

        ignored = False
        if current_tx:
            ignored = not nest

        else:
            ignored = False

        index = self.transaction_count() + 1

        return {
            "nest": nest,
            "name": name,
            "ignored": ignored,
            "index": index
        }

    def transaction_current(self):
        """Returns the current transaction dict, or empty dict if no tx"""
        return self.transactions[-1] if self.transactions else {}

    def transaction_exists(self):
        """return true if currently in a transaction

        this was previously named .in_transaction but it turns out SQLite has a
        property with that name
        """
        return self.transaction_count() > 0

    async def transaction_start(self, **kwargs):
        """start a transaction

        this will increment transaction semaphore and pass it to
        _transaction_start()
        """
        tx = self.transaction_info(**kwargs)
        self.transactions.push(tx)

        transaction_count = tx["index"]
        if tx["ignored"]:
            self.log_debug([
                f"{transaction_count}.",
                f"Ignoring {self} transaction {self.transaction_names()}",
            ])
            await self._transaction_ignoring(tx)

        else:
            self.log_debug([
                f"{transaction_count}.", 
                f"Start {self} transaction {self.transaction_names()}",
            ])
            if transaction_count == 1:
                await self._transaction_start()

            else:
                await self._transaction_starting(tx)

        return transaction_count

    async def transaction_stop(self):
        """stop/commit a transaction if ready"""
        transaction_count = self.transaction_count()
        if transaction_count > 0:
            tx = self.transactions.pop()
            if not tx["ignored"]:
                self.log_debug([
                    f"{transaction_count}.", 
                    f"Stopping {self} transaction",
                    self.transaction_names(tx),
                ])

                if transaction_count == 1:
                    await self._transaction_stop()

                else:
                    await self._transaction_stopping(tx)

        return self.transaction_count

    async def transaction_fail(self):
        """rollback a transaction if currently in one"""
        transaction_count = self.transaction_count()
        if transaction_count > 0:
            tx = self.transactions.pop()
            if not tx["ignored"]:
                self.log_debug([
                    f"{transaction_count}.",
                    f"Failing {self} transaction",
                    self.transaction_names(tx),
                ])

                if transaction_count == 1:
                    await self._transaction_fail()

                else:
                    await self._transaction_failing(tx)

    def __str__(self):
        return f"0x{id(self):02x}"


class InterfaceABC(LogMixin):
    """This is just a convenience abstract base class so child interfaces can
    easily see what methods they might need to implement. They should extend
    Interface and then implement the methods in this class
    """
    @classmethod
    async def configure(cls, config):
        """This is called by Config whenever it parses a DSN, it is meant to
        customize any configuration for the specific child interface
        """
        return config

    async def _connect(self, config):
        """See the docblock for Interface.connect to understand how the
        connection interface is used by Interface to make and manage connections
        """
        raise NotImplementedError()

    async def _configure_connection(self, **kwargs):
        """The wrapper method for this (Interface.configure_connection) is
        never called directly by Interface. It is up to the child interfaces
        to decide when a connection is ready and to call this method"""
        pass

    async def _free_connection(self, connection):
        """The wrapper method for this method is called at the end of the
        Interface.connection context manager"""
        pass

    async def _get_connection(self):
        """The wrapper method for this method is called at the beginning of the
        Interface.connection context manager"""
        raise NotImplementedError()

    async def _close(self):
        raise NotImplementedError()

    async def _readonly(self, readonly, **kwargs):
        raise NotImplementedError()

    async def _raw(self, query_str, *query_args, **kwargs):
        raise NotImplementedError()

    async def _set_table(self, schema, **kwargs):
        raise NotImplementedError()

    async def _get_tables(self, table_name, **kwargs):
        raise NotImplementedError()

    async def _delete_table(self, schema):
        raise NotImplementedError()

    async def _get_fields(self, table_name, **kwargs):
        raise NotImplementedError()

    async def _get_indexes(self, schema, **kwargs):
        raise NotImplementedError()

    async def _set_index(self, schema, name, field_names, **kwargs):
        raise NotImplementedError()

    async def _insert(self, schema, fields, **kwargs):
        raise NotImplementedError()

    async def _inserts(self, schema, field_names, field_values, **kwargs):
        raise NotImplementedError()

    async def _update(self, schema, fields, query, **kwargs):
        raise NotImplementedError()

    async def _upsert(
        self,
        schema,
        insert_fields,
        update_fields,
        conflict_field_names,
        **kwargs
    ):
        raise NotImplementedError()

    async def _delete(self, schema, query, **kwargs):
        raise NotImplementedError()

    async def _get(self, schema, query, **kwargs):
        raise NotImplementedError()

    async def _count(self, schema, query, **kwargs):
        raise NotImplementedError()

    async def _handle_unique_error(self, e, **kwargs):
        return False

    async def _handle_field_error(self, e, **kwargs):
        return False

    async def _handle_table_error(self, e, **kwargs):
        return False

    async def _handle_general_error(self, e, **kwargs):
        return False

    async def _handle_close_error(self, e, **kwargs):
        await self.close()
        return True

    def render(self, schema, query, **kwargs):
        """Render the query in a way that the interface can interpret it

        so in a SQL interface, this would render SQL, this is mainly used for
        debugging

        :param query: Query, the Query instance to render
        :param **kwargs: any named arguments
        :returns: Any
        """
        raise NotImplementedError()


class Interface(InterfaceABC):

    connected = False
    """true if a connection has been established, false otherwise"""

    config = None
    """a config.Connection() instance"""

    def __init__(self, config=None):
        self.config = config

    async def connect(self, config=None, *args, **kwargs):
        """connect to the interface

        The parent Interface (this class) never interacts with a connection
        directly, it uses the connection interface for all use/manipulation
        of the connection:

            * connect: This handles making the actual connection
            * configure_connection: This is called immediately after a 
                connection is ready and can be used to further configure the
                connection, if needed. This is *NEVER* called by this class, so
                it is up to the child interface to decide when a connection is
                ready and to call this method
            * close: This handles closing the actual connection
            * get_connection: Whenever a connection is needed this class will
                use the .connection context manager which calls this method to
                get an actual connection, this should return a connection
                instance
            * free_connection: Whenever this class is done with a connection
                this method will be called, usually at the end of the
                .connection context manager call
            * readonly: Responsible for setting a connection to readonly, this
                is called in .configure_connection but can be called separately
                also
            * is_connected: Returns True if an active connection has been made
            * reconnect: Close any open connections and then call .connect()
                again
            * connection: The connection context manager that calls
                .get_connection and .free_connection, most interactions with
                a connection will use this
            * transaction: wraps the .connection context manager and starts a
                transaction

        The reason why this class uses the connection interface and that no
        connection anything is fleshed out is because that allows the child
        interfaces a lot of flexibility in how they handle the connection, they
        can have a single connection or a pool of connections, etc.

        The amount of methods that will need to be actually implemented by the
        child interfaces is up to the child interfaces themselves

        :param config: Config, this doesn't need to be passed in because it can
            also be passed into the __init__ method, a valid Config object needs
            to be present before a connection can be made though
        :param *args: anything you want that will help the db connect
        :param **kwargs: anything you want that the backend db connection will
            need to actually connect
        """
        if self.connected:
            return self.connected

        if config:
            self.config = config

        await self.configure(self.config)

        try:
            await self._connect(self.config)
            self.connected = True

        except Exception as e:
            self.connected = False
            await self.raise_error(e)

        self.log_debug("Connected {}", self.config.interface_name)
        return self.connected

    async def reconnect(self):
        await self.close()
        await self.connect()

    async def configure_connection(self, **kwargs):
        """Configure a ready connection

        It is up to the child interfaces to decide when to call this method

        :param **kwargs:
            - connection: Any, the connection to configure
        """
        await self.execute(
            self._configure_connection,
            **kwargs
        )

        self.config.readonly = kwargs.get("readonly", self.config.readonly)

        # by default we can read/write, so only bother to run this if we
        # need to actually make the connection readonly
        if self.config.readonly:
            await self.readonly(
                self.config.readonly,
                **kwargs
            )

    async def get_connection(self):
        """Any time you need a connection it should be retrieved through
        .connection, and that method uses this method

        :returns: Connection instance
        """
        if not self.is_connected():
            await self.connect()

        connection = await self._get_connection()

        if connection.closed:
            # we've gotten into a bad state so let's try reconnecting
            await self.reconnect()
            connection = await self._get_connection()

        connection.interface = self

        self.log_debug(
            "Getting {} connection {}",
            self.config.interface_name,
            connection
        )
        return connection

    async def free_connection(self, connection):
        """When .connection is done with a connection it calls this method"""
        if self.is_connected():
            self.log_debug(
                "Freeing {} connection {}",
                self.config.interface_name,
                connection
            )
            await self._free_connection(connection)

    def is_connected(self):
        """Returns True if this Interface has been connected"""
        return self.connected

    async def close(self):
        """close an open connection"""
        if not self.connected:
            return True

        await self._close()
        self.connected = False
        self.log_debug("Closed Connection {}", self.config.interface_name)
        return True

    @asynccontextmanager
    async def connection(self, connection=None, **kwargs):
        """Any time you need a connection you should use this context manager,
        this is the only place that wraps exceptions in InterfaceError, so all
        connections should go through this method or .transaction if you need
        to start a transaction

        :Example:
            with self.connection(**kwargs) as connection:
                # do something with connection
        """
        free_connection = False

        prefix = kwargs.get("prefix", "")
        if prefix:
            prefix += " "

        try:
            if connection:
                if connection.closed:
                    self.log_warning(
                        "{}Existing connection {} is closed",
                        prefix,
                        connection
                    )

                    connection = None

                else:
                    self.log_debug(
                        "{}Using existing connection {}",
                        prefix,
                        connection
                    )

            if connection is None:
                free_connection = True

                self.log_debug(
                    "{}Getting connection",
                    prefix,
                )

                connection = await self.get_connection()

                self.log_debug(
                    "{}Got connection {}",
                    prefix,
                    connection
                )

            yield connection

        except Exception as e:
            await self.raise_error(e)

        finally:
            if free_connection and connection:
                await self.free_connection(connection)

                self.log_debug(
                    "{}Connection {} was freed",
                    prefix,
                    connection
                )

            else:
                self.log_debug(
                    "{}Existing connection {} was NOT freed",
                    prefix,
                    connection
                )

    @asynccontextmanager
    async def transaction(self, connection=None, **kwargs):
        """A simple context manager useful for when you want to wrap a bunch of
        db calls in a transaction, this is used internally for any write
        statements

        NOTE -- psycopg3 now has a transaction context manager:
            https://www.psycopg.org/psycopg3/docs/api/connections.html#psycopg.Connection.transaction

            It might be worth switching over to it sometime in the future? I'm
            not sure how that would work with sqlite though

        :Example:
            with self.transaction() as connection
                # do a bunch of calls
            # those db calls will be committed by this line
        """
        async with self.connection(connection, **kwargs) as connection:
            await connection.transaction_start(**kwargs)
            try:
                yield connection

            except Exception:
                await connection.transaction_fail()
                raise

            else:
                await connection.transaction_stop()

    async def readonly(self, readonly=True, **kwargs):
        """Make the connection read only (pass in True) or read/write (pass in
        False)

        :param readonly: boolean, True if this connection should be readonly,
            False if the connection should be read/write
        """
        self.log_warning([
            f"Setting interface {self.config.interface_name}",
            f"to readonly={readonly}",
        ])
        self.config.readonly = readonly

        await self.execute(
            self._readonly,
            readonly,
            **kwargs
        )

    async def execute_write(self, callback, *args, **kwargs):
        """Any write statements will use this method

        collectively "write statements":
            CREATE, DELETE, DROP, INSERT, or UPDATE
        """
        kwargs.setdefault("execute_in_transaction", True)
        return await self.execute(callback, *args, **kwargs)

    async def execute_read(self, callback, *args, **kwargs):
        """Any read statements will use this method

        collectively "read statements":
            SELECT
        """
        kwargs["prefix"] = "execute_read"
        # we need the connection so we can decide if we need to run the query
        # in a tx or not
        async with self.connection(**kwargs) as connection:
            kwargs["connection"] = connection

            kwargs.setdefault(
                "execute_in_transaction",
                connection.transaction_exists()
            )

            return await self.execute(callback, *args, **kwargs)

    async def execute(self, callback, *args, **kwargs):
        """Internal method. Execute the callback with args and kwargs, retrying
        the query if an error is raised that it thinks it successfully handled

        This is called by .execute_write, .execute_read, and .raw

        :param callback: callable, this will be run at-most twice
        :param *args: passed directly to callback as *args
        :param **kwargs: passed to callback as **kwargs, can have values added
        :returns: Any, whatever the callback returns
        """
        # we want to override the prefix at this point
        kwargs.pop("prefix", None)
        prefix = callback.__name__

        try:
            return await self._execute(
                callback,
                *args,
                prefix=prefix,
                **kwargs
            )

        except Exception as e:
            if await self.handle_error(e=e, prefix=prefix, **kwargs):
                return await self._execute(
                    callback,
                    *args,
                    prefix=f"{prefix}_retry",
                    **kwargs
                )

            else:
                await self.raise_error(e)

    async def _execute(self, callback, *args, **kwargs):
        """Internal method for .execute, this should never be called directly,
        it's broken out from .execute so .execute can re-run it if an error is
        encountered. Because the transaction is started in this method then
        it will be rolledback before .execute starts error handling, which is
        what we want.

        :param callback: callable, this will be run at-most twice
        :param *args: passed directly to callback as *args
        :param **kwargs: passed to callback as **kwargs, can have values added
            - execute_only: bool, don't wrap the callback call in a transaction
                or connection context manager
            - execute_in_transaction: bool, True if callback should be wrapped
                in a tx, only applies if execute_only=False
        :returns: Any, whatever the callback returns
        """
        execute_only = kwargs.get("execute_only", False)

        if execute_only:
            return await callback(*args, **kwargs)

        else:
            in_transaction = kwargs.get("execute_in_transaction", False)

            if in_transaction:
                async with self.transaction(**kwargs) as connection:
                    kwargs["connection"] = connection
                    return await callback(*args, **kwargs)

            else:
                async with self.connection(**kwargs) as connection:
                    kwargs["connection"] = connection
                    return await callback(*args, **kwargs)

    async def has_table(self, table_name, **kwargs):
        """Check to see if a table is in the db

        :param table_name: str, the table to check
        :returns: bool, True if the table exists, false otherwise
        """
        tables = await self.execute_read(
            self._get_tables,
            table_name,
            **kwargs
        )
        return len(tables) > 0

    async def get_tables(self, table_name="", **kwargs):
        """Get all the tables of the currently connected db

        :param table_name: str, if you would like to filter the tables list to
            only include matches with this name
        :returns: list, a list of table names
        """
        return await self.execute_read(
            self._get_tables,
            str(table_name),
            **kwargs
        )

    async def set_table(self, schema, **kwargs):
        """
        add the table to the db

        :param schema: Schema instance, contains all the information about the
            table
        """
        kwargs["prefix"] = "set_table"
        async with self.transaction(**kwargs) as connection:
            kwargs["connection"] = connection

            await self.execute_write(
                self._set_table,
                schema=schema,
                #connection=connection,
                #execute_only=True,
                **kwargs
            )

            for index_name, index in schema.indexes.items():
                await self.execute_write(
                    self._set_index,
                    schema=schema,
                    name=index.name,
                    field_names=index.field_names,
                    #connection=connection,
                    #execute_only=True
                    **index.options,
                )

    async def unsafe_delete_table(self, schema, **kwargs):
        """wrapper around delete_table that matches the *_tables variant and
        denotes that this is a serious operation

        remove a table matching schema from the db

        :param schema: Schema instance, the table to delete
        """
        await self.execute_write(
            self._delete_table,
            schema=schema,
            **kwargs
        )
        return True

    async def unsafe_delete_tables(self, **kwargs):
        """Removes all the tables from the db

        https://github.com/Jaymon/prom/issues/75
        """
        kwargs["prefix"] = "unsafe_delete_tables"
        async with self.transaction(**kwargs) as connection:
            kwargs['connection'] = connection
            kwargs.setdefault('nest', False)
            for table_name in await self.get_tables(**kwargs):
                # we don't wrap this in an .execute_write because there isn't
                # anything to recover from if it fails
                await self._delete_table(table_name, **kwargs)
        return True

    async def get_indexes(self, schema, **kwargs):
        """
        get all the indexes

        schema -- Schema()

        return -- dict -- the indexes in {indexname: fields} format
        """
        return await self.execute_read(
            self._get_indexes,
            schema=schema,
            **kwargs
        )

    async def set_index(self, schema, name, field_names, **kwargs):
        """
        add an index to the table

        :param schema: Schema
        :param name: str, the name of the index
        :param field_names: list, the fields the index should be on
        :param **index_options: any index options that might be useful to
            create the index
        """
        await self.execute_write(
            self._set_index,
            schema=schema,
            name=name,
            field_names=field_names,
            **kwargs,
        )
        return True

    async def insert(self, schema, fields, **kwargs):
        """Persist fields into the db

        :param schema: Schema instance, the table the query will run against
        :param fields: dict, the fields {field_name: field_value} to persist
        :param **kwargs: passed through
        :returns: mixed, will return the primary key values
        """
        return await self.execute_write(
            self._insert,
            schema=schema,
            fields=fields,
            **kwargs
        )

    async def inserts(self, schema, field_names, field_rows, **kwargs):
        """Persist the field names found in all the field rows

        :param schema: Schema
        :param field_names: list, the field names that will be checked for each
            row in field_rows and used to turn each dict in field_rows to a
            tuple
        :param field_rows: Sequence, if an iterator of dict instances with keys
            found in field_names, if a key is missing it will have None set as
            the value, if it's an iterator of values then the tuple value
            ordering should match with field_names
        :param **kwargs: passed through
        :returns: bool, True if the query executed successfully
        """
        def field_values(field_names, field_rows):
            for fields in field_rows:
                if isinstance(fields, Mapping):
                    row = []
                    for field_name in field_names:
                        row.append(fields.get(field_name, None))
                    yield row

                else:
                    yield fields

        await self.execute_write(
            self._inserts,
            schema=schema,
            field_names=field_names,
            field_values=field_values(field_names, field_rows),
            **kwargs,
        )
        return True

    async def update(self, schema, fields, query, **kwargs):
        """Persist the query.fields into the db that match query.fields_where

        :param schema: Schema instance, the table the query will run against
        :param fields: dict, the fields {field_name: field_value} to persist
        :param query: Query instance, will be used to create the where clause
        :param **kwargs:
            * count_result: bool, True if you want to return how many rows
                where touched by the update query
            * ignore_result: bool, True if you don't care about returning
                any result
            * ignore_return_clause: bool, True if you don't want to return
                the updated values of the rows that were touched by the query
        :returns: list[dict]|int|bool, by default this will return the touched
            rows that were updated by the query. If return value is an int then
            it will be how many rows were updated. If return value is a boolean
            then it will be True showing that the query succeeded
        """
        return await self.execute_write(
            self._update,
            schema=schema,
            fields=fields,
            query=query,
            **kwargs,
        )

    async def upsert(
        self,
        schema,
        insert_fields,
        update_fields,
        conflict_field_names,
        **kwargs
    ):
        """Perform an upsert (insert or update) on the table

        :param schema: Schema instance, the table the query will run against
        :param insert_fields: dict, these are the fields that will be inserted
        :param update_fields: dict, on a conflict with the insert_fields, these
            fields will instead be used to update the row
        :param conflict_field_names: list, the field names that will decide if
            an insert or update is performed
        :param **kwargs: anything else
        :returns: str|int|None, the primary key
        """
        return await self.execute_write(
            self._upsert,
            schema=schema,
            insert_fields=insert_fields,
            update_fields=update_fields,
            conflict_field_names=conflict_field_names,
            **kwargs,
        )

    async def delete(self, schema, query, **kwargs):
        """delete matching rows according to query filter criteria

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria, this will fail if
            empty
        :returns: int, how many rows were deleted ... I think
        """
        if not query or not query.fields_where:
            raise ValueError("aborting delete because there is no where clause")

        return await self.unsafe_delete(schema, query, **kwargs)

    async def unsafe_delete(self, schema, query, **kwargs):
        """delete matching rows

        WARNING -- this can clear the whole table, you should mainly use
            .delete and always include filtering criteria in the query

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria, this will fail if
            empty
        :returns: int, how many rows were deleted ... I think
        """
        return await self.execute_write(
            self._delete,
            schema=schema,
            query=query,
            **kwargs
        )

    async def raw(self, query_str, *query_args, **kwargs):
        """
        run a raw query on the db

        :param query_str: str, the query to run
        :param *query_args: if the query_str is a formatting string, pass the
            values in this
        :param **kwargs: any query options can be passed in by using key=val
            syntax
        """
        return await self.execute(
            self._raw,
            query_str,
            *query_args,
            **kwargs
        )

    async def get_fields(self, table_name, **kwargs):
        return await self.execute_read(
            self._get_fields,
            str(table_name),
            **kwargs
        )

    async def one(self, schema, query=None, **kwargs):
        """get one row from the db matching filters set in query

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria
        :return: dict, the matching row
        """
        return await self.get(schema, query, fetchone=True, **kwargs) or {}

    async def get(self, schema, query=None, **kwargs):
        """get matching rows from the db matching filters set in query

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria
        :returns: list, a list of matching dicts
        """
        ret = await self.execute_read(
            self._get,
            schema=schema,
            query=query or Query(),
            **kwargs
        )
        return ret or []

    async def count(self, schema, query=None, **kwargs):
        """count matching rows according to query filter criteria

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria
        :returns: list, a list of matching dicts
        """
        ret = await self.execute_read(
            self._count,
            schema=schema,
            query=query or Query(),
            **kwargs
        )
        return int(ret) if ret else 0

    async def handle_error(self, e, **kwargs):
        """Try and handle the error, return False if the error can't be handled

        :param e: Exception, the caught exception
        :param **kwargs:
            - schema: Schema, this does not have to be there, but usually is
        :returns: bool, True if the error was handled, False if it wasn't
        """
        ret = False

        prefix = kwargs.get("prefix", "")
        self.log_warning(["Handling", prefix, f"error: {e}"])

        e = self.create_error(e)

        if isinstance(e, CloseError):
            self.log_debug("Handling a close error")
            ret = await self._handle_close_error(e=e, **kwargs)

        else:
            kwargs["prefix"] = f"{prefix}_handle_error"
            async with self.transaction(**kwargs) as connection:
                kwargs["connection"] = connection

                if isinstance(e, UniqueError):
                    self.log_debug("Handling a unique error")
                    ret = await self._handle_unique_error(e=e, **kwargs)

                elif isinstance(e, FieldError):
                    self.log_debug("Handling a field error")
                    ret = await self._handle_field_error(e=e, **kwargs)

                elif isinstance(e, TableError):
                    self.log_debug("Handling a table error")
                    ret = await self._handle_table_error(e=e, **kwargs)

                else:
                    self.log_debug("Handling a general error")
                    ret = await self._handle_general_error(e=e, **kwargs)

        if ret:
            self.log_info(["Successfully handled", prefix, "error"])

        else:
            self.log_warning(["Failed to handle", prefix, "error"])

        return ret

    def raise_error(self, e, **kwargs):
        """raises e

        :param e: Exception, if a built-in exception then it's raised, if any
            other error then it will be wrapped in an InterfaceError
        """
        e2 = self.create_error(e, **kwargs)
        if e2 is not e:
            raise e2 from e
        else:
            raise e

    def create_error(self, e, **kwargs):
        """create the error that you want to raise, this gives you an
        opportunity to customize the error

        allow python's built in errors to filter up through
        https://docs.python.org/2/library/exceptions.html

        :param e: Exception, this error will be wrapped in an InterfaceError (or
            whatever is in kwargs["error_class"] if it isn't already an instance
            of that class
        :param **kwargs:
            - error_class: InterfaceError
            - error_module: module, the dbapi module
        """
        error_class = kwargs.get("error_class", InterfaceError)
        if not isinstance(e, error_class):
            if not hasattr(builtins, e.__class__.__name__):
                if "error_module" in kwargs:
                    errmod_name = kwargs["error_module"].__name__
                    if errmod_name in e.__class__.__module__:
                        e = error_class(e)

                else:
                    e = error_class(e)

        return e




# -*- coding: utf-8 -*-
import os
import random
import string
import datetime
import decimal

# make this module act like global testdata
from testdata import *
from testdata import __getattr__

from prom.compat import *
from prom.query import Query
from prom.model import Orm
from prom.config import Schema, DsnConnection, Field, Index, AutoIncrement
from prom.interface import find_environ
from prom.utils import make_dict


class InterfaceData(TestData):
    """Hold all helper methods to create db tables and get tests ready

    These testdata methods are specifically to test prom itself and should
    never be used outside of prom's tests
    """

    interfaces = set()
    """Every interface that is created gets added to this set so that it can be
    closed"""

    def get_interfaces(self):
        """Return all currently configured interfaces in a list"""
        return list(self.interfaces)

    def get_interface(self, interface=None):
        """Get an interface, this does not connect the interface

        :param Interface: Interface, if you pass one in that will be returned
            instead of a new interface created
        :returns: Interface
        """
        return interface or self.create_interface()

    def create_interface(self):
        """Create an interface and return it, this does not connect the
        interface

        NOTE -- this uses a class property `interface_class` that needs to be
        set in a test setup method like TestCase.setUp or the like, if that
        class property is set then it will use that class to create the
        interface, otherwise it just creates the first found interface

        :return: Interface
        """
        if interface_class := getattr(self, "interface_class", None):
            return self.find_interface(interface_class)

        else:
            for inter in self.create_environ_interfaces():
                return inter

    def create_dsn_interface(self, dsn):
        """Create an interface with the given DSN

        :param dsn: str, the DSN string
        :returns: Interface
        """
        conn = DsnConnection(dsn)
        inter = conn.interface
        self.interfaces.add(inter)
        return inter

    def create_environ_connections(self, dsn_env_name="PROM_TEST_DSN"):
        """creates all the connections that are defined in the environment under
        <dsn_env_name>_N where N can be any integer

        NOTE -- since this can create multiple connections if the environment
        variables are postfixed with _N (eg, *_1, *_2) you can set the
        environment variable `PROM_TEST_DSN_INDEX` to an index to limit
        connection creation to only that index, so to limit to just *_1:

            export PROM_TEST_DSN_INDEX=1

        :param dsn_env_name: str, the name of the environment variable that 
            should be read to create the connections
        :returns: generator[Connection]
        """
        found = False
        if dsn_index := os.environ.get("PROM_TEST_DSN_INDEX", 0):
            for conn in find_environ(f"{dsn_env_name}_{dsn_index}"):
                found = True
                yield conn

        else:
            for conn in find_environ(dsn_env_name):
                found = True
                yield conn

        if not found:
            raise ValueError("No connection found, set PROM_TEST_DSN")

    def create_environ_interfaces(self):
        """find any interfaces that match dsn_env_name and yield them

        This is basically .create_environ_connections but wraps the connection
        instance in an Interface instance

        :returns: generator[Interface]
        """
        for conn in self.create_environ_connections():
            inter = conn.interface
            self.interfaces.add(inter)
            yield inter

    def get_environ_interface_classes(self):
        """Returns all the Interface classes of the given environment

        :returns: generator[type], when you don't want the Connection instance,
            you don't want an Interface instance, you just want the Connection
            class
        """
        for conn in self.create_environ_connections():
            yield conn.interface_class

    def find_interface(self, interface_class):
        """Find an Interface that matches `interface_class`

        :returns: Interface
        """
        for inter in self.create_environ_interfaces():
            if isinstance(inter, interface_class):
                return inter

    def get_table_name(self, table_name=None, prefix=""):
        """return a random table name

        :param table_name: str, if passed in then this will be returned
        :param prefix: str, if passed in then a table name will be randomly
            generated that starts with this
        :returns: str, a usable table name
        """
        if table_name:
            return table_name

        return "{}{}_table".format(
            prefix,
            "".join(
                random.sample(string.ascii_lowercase, random.randint(5, 15))
            )
        )

    def get_orm_class(self, table_name=None, prefix="orm_class", **properties):
        """Return an orm class

        :param table_name: str, the table the orm class will wrap
        :param prefix: str, the class name will start with this
        :param **properties: these will be set on the orm class, if they are
            Field instances then those will be set as schema fields, if they
            are Index instances they will be schema indexes
            - interface: Interface, the interface the returned class will use
            - connection_name: str, an interface will be created with this
                connection name
        :returns: type, an Orm child class
        """
        tn = self.get_table_name(table_name, prefix=prefix)
        parent_class = properties.get("parent_class", Orm)

        properties["table_name"] = tn

        if "interface" not in properties:
            if "connection_name" not in properties:
                properties["interface"] = self.get_interface()

        has_field = False
        for k, v in properties.items():
            if isinstance(v, Field):
                has_field = True
                break

            elif isinstance(v, type) and issubclass(v, Field):
                has_field = True
                break

        if not has_field:
            properties.update({
                "foo": Field(int, True),
                "bar": Field(str, True),
                "ifoobar": Index("foo", "bar"),
            })

        orm_class = type(
            String(tn),
            (parent_class,),
            properties,
        )

        return orm_class

    def get_orm(self, table_name=None, prefix="orm", **fields):
        """Returns an Orm instance

        :param table_name: passed through to `.get_orm_class`
        :param prefix: passed through to `.get_orm_class`
        :param **fields: these are used to instantiate an instance of the
            created orm class
        :returns: Orm
        """
        orm_class = self.get_orm_class(table_name, prefix=prefix)
        t = orm_class(**fields)
        return t

    async def create_orms(self, table_name=None, count=0, **fields):
        """Create count orms at table_name with fields

        see `.get_orm_class` for the arguments

        :param table_name: passed through to `.get_orm_class`
        :param prefix: passed through to `.get_orm_class`
        :param **fields: passed through to `.get_orm_class`
        :param count: int, how many orms you want created
        :returns: type, the Orm class, the Orm class created with table_name
        """
        orm_class = self.get_orm_class(table_name, **fields)
        count = count or self.get_int(1, 10)
        await self.insert(orm_class, count)
        return orm_class

    async def create_orm(self, table_name=None, **fields):
        """Creat an orm with the given field values

        :param table_name: passed through to `.get_orm`
        :param prefix: passed through to `.get_orm`
        :param **fields: passed through to `.get_orm`
        :returns: Orm, this will be persisted in the db
        """
        orm_class = self.get_orm(table_name, **fields)
        fs = self.get_fields(orm_class.schema)
        return await orm_class.create(fs)

    def get_table(self, interface=None, table_name=None, **fields_or_indexes):
        """Return an interface and schema for a table in the db

        This does not try and create anything

        :param interface: Interface, the interface you want to use
        :param table_name: str, the table name, uses .get_table_name() if None
        :param **fields_or_indexes: keys will be field name and the values
            should be Field or Index instances
        :returns: tuple[Interface, Schema]
        """
        interface = self.get_interface(interface)
        schema = self.get_schema(table_name, **fields_or_indexes)
        return interface, schema

    async def create_table(self, *args, **kwargs):
        """Create a table on the db

        :param *args: passed through to `.get_table`
        :param **kwargs: passed through to `.get_table`
        :returns: tuple[Interface, Schema]
        """
        interface, schema = self.get_table(*args, **kwargs)
        await interface.set_table(schema)
        return interface, schema

    def get_schema(self, table_name=None, prefix="schema", **fields_or_indexes):
        """Get a Schema instance

        :param table_name: str, the table name you want the schema to have, if
            not given it will be automatically created using `prefix`
        :param prefix`: str, used to create the table name
        :param **fields_or_indexes: fields will be Field instances with the key
            being the name. Indexes will be Index instances with key as the
            index name
        :returns: Schema
        """
        if not fields_or_indexes:
            fields_or_indexes.setdefault("foo", Field(int, True))
            fields_or_indexes.setdefault("bar", Field(str, True))
            fields_or_indexes.setdefault("ifoobar", Index("foo", "bar"))

        fields_or_indexes.setdefault("_id", AutoIncrement())

        # remove any None values
        for k in list(fields_or_indexes.keys()):
            if not fields_or_indexes[k]:
                fields_or_indexes.pop(k)

        s = Schema(
            self.get_table_name(table_name, prefix=prefix),
            **fields_or_indexes
        )
        return s

    def get_schema_all(self, inter=None):
        """return a schema that has a field for all supported standard field
        types"""

        # this is for foreign key fields
        orm_class = self.get_orm_class()
        if inter:
            orm_class.interface = inter
            orm_class.install()

        s = Schema(
            self.get_table_name(),
            _id=AutoIncrement(),
            a_bool_y=Field(bool, True),
            a_bool_n=Field(bool, False),
            a_sint_y=Field(int, True, size=50),
            a_sint_n=Field(int, False, size=50),
            a_dec_y=Field(decimal.Decimal, True),
            a_dec_n=Field(decimal.Decimal, False),
            a_float_y=Field(float, True, size=10),
            a_float_n=Field(float, False, size=10),
            a_long_y=Field(long, True),
            a_long_n=Field(long, False),
            a_fk_y=Field(orm_class, True),
            a_fk_n=Field(orm_class, False),
            a_dt_y=Field(datetime.datetime, True),
            a_dt_n=Field(datetime.datetime, False),
            a_d_y=Field(datetime.date, True),
            a_d_n=Field(datetime.date, False),
            a_int_y=Field(int, True),
            a_int_n=Field(int, False),
            a_str_y=Field(str, True),
            a_str_n=Field(str, False),
            a_vchar_y=Field(str, True, max_size=512),
            a_vchar_n=Field(str, False, max_size=512),
            a_char_y=Field(str, True, size=32),
            a_char_n=Field(str, False, size=32),
        )

        return s

    def get_query(self, table_name=None, prefix="query", **properties):
        """Get a query instance

        :param table_name: str, if passed in the table will be named this
        :param prefix: str, a random table name will be created using this
            prefix
        :returns: Query, this query will belong to an Orm, so Query.orm_class
            should return the orm class
        """
        orm_class = self.get_orm_class(
            table_name,
            prefix=prefix,
            **properties
        )
        return orm_class.query

    def get_fields(self, schema, **field_kwargs):
        """return the fields of orm with randomized data

        :param schema: Schema, values for all the fields in this schema will be
            randomly generated
        :param **field_kwargs: specific field values you want the returned dict
            to have
        :returns: dict[str, Any]
        """
        fields = {}
        for k, v in schema.fields.items():
            if v.is_pk(): continue
            if v.is_ref(): continue

            if issubclass(v.interface_type, basestring):
                fields[k] = self.get_words()

            elif issubclass(v.interface_type, bool):
                fields[k] = True if random.randint(0, 1) == 1 else False

            elif issubclass(v.interface_type, int):
                fields[k] = self.get_int32()

            elif issubclass(v.interface_type, datetime.datetime):
                fields[k] = self.get_past_datetime()

            elif issubclass(v.interface_type, float):
                fields[k] = self.get_float()

            else:
                raise ValueError("{}".format(v.interface_type))

        fields.update(field_kwargs)
        return fields

    def get_insert_fields(self, *args, **kwargs):
        """Gets everything needed to insert fields into a table

        Internal method used by the other .insert* methods

        :param *args:
            - interface, schema, count
            - query, count
            - (interface, schema), count
            - orm_class, count
            - orm instance, count
            - orm_class
            - orm instance
        :returns: tuple[Interface, Schema, Orm, list[dict]], the list will be
            count dicts with fields to insert using Schema
        """
        if len(args) == 1:
            o = args[0]
            count = 1

        elif len(args) == 2:
            o = args[0]
            count = args[1]

        elif len(args) == 3:
            o = (args[0], args[1])
            count = args[2]

        else:
            raise ValueError(f"Passed in {len(args)} arguments")

        if isinstance(o, Query):
            orm_class = o.orm_class
            interface = o.interface
            schema = orm_class.schema

        elif isinstance(o, tuple):
            orm_class = None
            interface, schema = o

        elif isinstance(o, Orm):
            orm_class = type(o)
            schema = orm_class.schema
            interface = o.interface

        elif issubclass(o, Orm):
            orm_class = o
            schema = orm_class.schema
            interface = o.interface

        else:
            raise ValueError("couldn't get fields for object {}".format(o))

        fields_list = []
        for i in range(count):
            fields_list.append(self.get_fields(schema, **kwargs))

        return interface, schema, orm_class, fields_list

    async def insert(self, *args, **kwargs):
        """This has the same signature as .get_insert_fields but inserts the
        generated fields into the db

        :param *args: passed through to .get_insert_fields
        :param **kwargs: passed through to .get_insert_fields
        :returns: int|list, if only one row was inserted then this will return
            the pk of that row, if multiple rows were inserted it will return
            a list of primary keys for all the rows inserted
        """
        pks = []

        interface, schema, orm_class, fields = self.get_insert_fields(
            *args,
            **kwargs
        )
        for fs in fields:
            d = await interface.insert(schema, fs)
            if pk_name := schema.pk_name:
                pks.append(d[pk_name])

            else:
                pks.append(None)

        return pks if len(pks) > 1 else pks[0]

    async def insert_fields(self, o, fields=None, **fields_kwargs):
        """Insert fields

        :param o: Query|tuple[Interface, Schema]|Orm
        :param fields: dict|None
        :param **fields_kwargs: merged with fields
        :returns: Any, the primary key for the row containing the inserted
            fields
        """
        fields = make_dict(fields, fields_kwargs)
        return await self.insert(o, 1, **fields)

    async def insert_orm(self, orm_class, fields=None, **fields_kwargs):
        """Insert fields into orm_class and return an orm instance

        :param orm_class: Orm, the orm class
        :param fields: dict|None
        :param **fields_kwargs: merged with fields
        :returns: Orm, an instance of orm_class that represents a row in the db
        """
        pk = await self.insert_fields(orm_class, fields, **fields_kwargs)
        return await orm_class.query.eq_pk(pk).one()




# -*- coding: utf-8 -*-

import dsnparse

from ..config import DsnConnection


interfaces = {}
"""holds all the configured interfaces"""


def find_environ(dsn_env_name='PROM_DSN', connection_class=DsnConnection):
    """Returns Connection instances found in the environment"""
    return dsnparse.parse_environs(dsn_env_name, parse_class=connection_class)


def configure_environ(dsn_env_name='PROM_DSN', connection_class=DsnConnection):
    """
    configure interfaces based on environment variables

    by default, when prom is imported, it will look for PROM_DSN, and PROM_DSN_N
    (where N is 1 through infinity) in the environment, if it finds them, it
    will assume they are dsn urls that prom understands and will configure db
    connections with them. If you don't want this behavior (ie, you want to
    configure prom manually) then just make sure you don't have any environment
    variables with matching names

    The num checks (eg PROM_DSN_1, PROM_DSN_2) go in order, so you can't do
    PROM_DSN_1, PROM_DSN_3, because it will fail on _2 and move on, so make
    sure your num dsns are in order (eg, 1, 2, 3, ...)

    example --
        export PROM_DSN_1=some.Interface://host:port/dbname#i1
        export PROM_DSN_2=some.Interface://host2:port/dbname2#i2
        $ python
        >>> import prom
        >>> print prom.interfaces # prints a dict with interfaces i1 and i2 keys

    :param dsn_env_name: string, the name of the environment variables
    """
    inters = []
    for c in find_environ(dsn_env_name, connection_class=connection_class):
        inter = c.interface
        set_interface(inter, c.name)
        inters.append(inter)

    return inters


def configure(dsn, connection_class=DsnConnection):
    """
    configure an interface to be used to query a backend

    you use this function to configure an Interface using a dsn, then you can
    get that interface using the get_interface() method

    :param dsn: str, a properly formatted prom dsn, see DsnConnection for how
        to format the dsn
    :param connection_class: type, the class that will hold the parsed result
    """
    c = dsnparse.parse(dsn, parse_class=connection_class)
    inter = c.interface
    set_interface(inter, c.name)
    return inter


def get_interfaces():
    global interfaces

    if not interfaces:
        configure_environ()

    return interfaces


def set_interface(interface, name=''):
    """
    don't want to bother with a dsn? Use this method to make an interface
    available
    """
    global interfaces

    if not interface:
        raise ValueError('interface is empty')

    # close down the interface before we discard it
    if name in interfaces:
        pout.v(interfaces[name].is_connected())
        if interfaces[name].is_connected():
            raise ValueError("Cannot replace an open interface")

    interfaces[name] = interface


def get_interface(name=''):
    """
    get an interface that was created using configure()

    name -- string -- the name of the connection for the interface to return
    """
    global interfaces

    if not interfaces:
        configure_environ()

    return interfaces[name]

